// Scintilla source code edit control
// PlatfGDK.cc - implementation of platform facilities on GTK+/Linux
// Copyright 1998-1999 by Neil Hodgson <neilh@hare.net.au>
// The License.txt file describes the conditions under which this software may be distributed.

#include <windows.h>

#include "Platform.h"

Surface::Surface() {
	hdc = 0;
	pen = 0;
	brush = 0;
}

Surface::~Surface() {
	if (pen) {
		SelectObject(hdc, penOld);
		DeleteObject(pen);
		pen = 0;
		penOld = 0;
	}
	if (brush) {
		SelectObject(hdc, brushOld);
		DeleteObject(brush);
		brush = 0;
		brushOld = 0;
	}
	if (fontOld) {
		SelectObject(hdc, fontOld);
		fontOld = 0;
	}
}

void Surface::Init(HDC hdc_) {
	hdc = hdc_;
	pen = 0;
	penOld = 0;
	brush = 0;
	brushOld = 0;
	fontOld = 0;
	//?HBRUSH brush = CreateSolidBrush(back);
	//?HBRUSH brushOld = (HBRUSH)SelectObject(hdc, brush);
}

Surface::PenColor(COLORREF fore) {
	if (pen) {
		SelectObject(hdc, penOld);
		DeleteObject(pen);
		pen = 0;
		penOld = 0;
	}
	pen = CreatePen(0,1,fore);
	penOld = (HPEN)SelectObject(hdc, pen);
}

Surface::BrushColor(COLORREF back) {
	if (brush) {
		SelectObject(hdc, brushOld);
		DeleteObject(brush);
		brush = 0;
		brushOld = 0;
	}
	brush = CreateSolidBrush(back);
	brushOld = (HBRUSH)SelectObject(hdc, brush);
}

void Surface::MoveTo(int x_, int y_) {
	MoveToEx(hdc, x_, y_, 0);
}

void Surface::LineTo(int x_, int y_) {
	::LineTo(hdc, x_, y_);
}

void Surface::Polygon(POINT *pts, int npts, COLORREF fore,
                      COLORREF back) {
	PenColor(fore);
	BrushColor(back);
	::Polygon(hdc, pts, npts);
}

void Surface::Rectangle(RECT rc, COLORREF fore, COLORREF back) {
	PenColor(fore);
	BrushColor(back);
	::Rectangle(hdc, rc.left, rc.top, rc.right, rc.bottom);
}

void Surface::FillRectangle(RECT rc, COLORREF back) {
	//PenColor(pen);
	BrushColor(back);
	::FillRect(hdc, &rc, brush);
}

void Surface::RoundedRectangle(RECT rc, COLORREF fore, COLORREF back) {
	RoundRect(hdc,
          	rc.left + 1, rc.top,
          	rc.right - 1, rc.bottom,
          	8, 8 );
}

void Surface::Ellipse(RECT rc, COLORREF fore, COLORREF back) {
	PenColor(fore);
	BrushColor(back);
	::Ellipse(hdc, rc.left, rc.top, rc.right, rc.bottom);
}

void Surface::DrawText(RECT rc, HFONT font, int ybase, char *s, int len, COLORREF fore, COLORREF back) {
	if (fontOld) {
		SelectObject(hdc, fontOld);
		fontOld = 0;
	}
	fontOld = SelectObject(hdc, font);
	SetTextColor(hdc, fore);
	SetBkColor(hdc, back);
	ExtTextOut(hdc, rc.left, ybase, ETO_OPAQUE, &rc, s, len, NULL);
}

int Surface::WidthText(HFONT font, char *s, int len) {
	if (fontOld) {
		SelectObject(hdc, fontOld);
		fontOld = 0;
	}
	fontOld = SelectObject(hdc, font);
	SIZE sz;
	GetTextExtentPoint32(hdc, s, len, &sz);
	return sz.cx;
}

int Surface::WidthChar(HFONT font, char ch) {
	if (fontOld) {
		SelectObject(hdc, fontOld);
		fontOld = 0;
	}
	fontOld = SelectObject(hdc, font);
	SIZE sz;
	GetTextExtentPoint32(hdc, &ch, 1, &sz);
	return sz.cx;
}

