// Tide 0.90 - tiny IDE to debug Scintilla
// Tide.cc - main code for the tiny IDE
// Copyright 1998-1999 by Neil Hodgson <neilh@hare.net.au>
// The License.txt file describes the conditions under which this software may be distributed.

#ifndef GTK
#include <windows.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>

#ifdef GTK
#include <unistd.h>
#include <gtk/gtk.h>
#include <glib.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#define stricmp strcasecmp
#include "gdk/gdkkeysyms.h"
#include "WinDefs.h"
#else
#include <io.h>
#include <process.h>
#include "Scintilla.h"
#ifdef _MSC_VER
#include <direct.h>
#include <richedit.h>
#else
extern "C" LONG STDCALL timeGetTime();
#endif
#define LRGB RGB
#endif

#include "Tide.h"
#include "KeyWords.h"
#include "PropSet.h"
#include "Scintilla.h"

#ifdef unix
const char pathSepString[] = "/";
const char pathSepChar = '/';
#else
const char pathSepString[] = "\\";
const char pathSepChar = '\\';
#endif

#ifdef GTK

#define COLORREF GdkColor
#define POINT GdkPoint
#define HFONT GdkFont*
#define DWORD long
#define UINT unsigned int
#define LRESULT long
#define BOOL bool

#define MAX_PATH 260

static void destroyDialog(GtkWidget *dialog) {
	gtk_grab_remove(dialog);
}

static GtkWidget *messageBoxDialog = 0;
static int messageBoxResult = 0;

static void messageBoxOK(GtkWidget *w, gpointer p) {
	gtk_widget_destroy(GTK_WIDGET(messageBoxDialog));
	messageBoxDialog = 0;
	messageBoxResult = reinterpret_cast<int>(p);
}

static GtkWidget *AddMBButton(GtkWidget *dialog, const char *label,
                          	int val, bool isDefault=false) {
	GtkWidget *button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
                   	GtkSignalFunc(messageBoxOK), reinterpret_cast<gpointer>(val));
	if (isDefault) {
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
	}
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
                   	button, TRUE, TRUE, 0);
	if (isDefault) {
		gtk_widget_grab_default(GTK_WIDGET(button));
	}
	gtk_widget_show(button);
	return button;
}

int MessageBox(GtkWidget *,char *m,char *t="Tide",int style=MB_OK) {
	if (!messageBoxDialog) {
		messageBoxResult = -1;
		messageBoxDialog = gtk_dialog_new();
		gtk_window_set_title(GTK_WINDOW(messageBoxDialog), 0);
		gtk_container_border_width(GTK_CONTAINER(messageBoxDialog), 0);

		gtk_signal_connect(GTK_OBJECT(messageBoxDialog),
                   		"destroy", GtkSignalFunc(destroyDialog), 0);

		if (style == MB_OK) {
			AddMBButton(messageBoxDialog, "OK", IDOK, true);
		} else {
			AddMBButton(messageBoxDialog, "Yes", IDYES, true);
			AddMBButton(messageBoxDialog, "No", IDNO);
			if (style == MB_YESNOCANCEL) {
				AddMBButton(messageBoxDialog, "Cancel", IDCANCEL);
			}
		}

		GtkWidget *label = gtk_label_new(m);
		gtk_misc_set_padding(GTK_MISC(label), 10, 10);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(messageBoxDialog)->vbox),
                   		label, TRUE, TRUE, 0);
		gtk_widget_show(label);

		//gtk_widget_grab_focus(GTK_WIDGET(defaultButton));
		//gtk_window_set_modal(GTK_WINDOW(messageBoxDialog), TRUE);  /* This line HURTS!!!! */

		gtk_widget_show(messageBoxDialog);

		gtk_grab_add(messageBoxDialog);

		//dprintf("MSGBOX: [%s] %s\n", t, m);

		while (messageBoxResult < 0) {
			gtk_main_iteration();
		}

	}
	return messageBoxResult;
}

struct RECT {
	int left;
	int top;
	int right;
	int bottom;
};

static COLORREF RGB(unsigned int red, unsigned int green, unsigned int blue) {
	GdkColor ret;
	ret.red = red * (65535/255);
	ret.green = green * (65535/255);
	ret.blue = blue * (65535/255);
	// the pixel value indicates the index in the colormap of the color.
	// it is simply a combination of the RGB values we set earlier
	ret.pixel = (gulong)(red*65536 + green*256 + blue);
	return ret;
}

static LONG LRGB(unsigned int red, unsigned int green, unsigned int blue) {
	return (red*65536 + green*256 + blue);
}

DWORD timeGetTime() {
	return 0;
}

static void GetClientRect(HWND win, RECT *rc) {
	rc->left = 0;
	rc->top = 0;
	rc->right = win->allocation.width;
	rc->bottom = win->allocation.height;
}
static void SetWindowPos(HWND hwnd, HWND wRelative, int left, int top,
                     	int w, int h, int flags) {
}
#else
HWND currentDlg = 0;
// AddStyledText only called from DefaultDlg so static size buffer is OK
void AddStyledText(HWND hwnd, const char *s, int attr) {
	char buf[1000];
	int len = strlen(s);
	for (int i=0;i<len;i++) {
		buf[i*2] = s[i];
		buf[i*2+1] = attr;
	}
	SendMessage(hwnd, SCI_ADDSTYLEDTEXT, len*2, reinterpret_cast<LPARAM>(buf));
}

// DefaultDlg,DoDialog is a bit like something in PC Magazine May 28, 1991, page 357
// DefaultDlg is only used for about box
int PASCAL DefaultDlg(HWND hDlg,WORD message,WPARAM wParam,LONG) {
	switch (message) {

	case WM_INITDIALOG: {
			HWND wsci = GetDlgItem(hDlg, IDABOUTSCINTILLA);
			if (wsci) {
				SendMessage(wsci, SCI_SETBACK, RGB(0,0,0), 0);
				SendMessage(wsci, EM_SETREADONLY, 0, 0);
				AddStyledText(wsci, "Tide\n", 0);
				SendMessage(wsci, SCI_STYLESETFORE, 0, RGB(0xff,0xff,0xff));
				SendMessage(wsci, SCI_STYLESETBACK, 0, RGB(0,0,0x80));
				SendMessage(wsci, SCI_STYLESETSIZE, 0, 22);
				AddStyledText(wsci, "Version 0.90\n", 1);
				SendMessage(wsci, SCI_STYLESETFORE, 1, RGB(0xff,0xff,0xff));
				SendMessage(wsci, SCI_STYLESETBACK, 1, RGB(0,0,0x80));
				SendMessage(wsci, SCI_STYLESETSIZE, 1, 12);
				AddStyledText(wsci, "by Neil Hodgson.\n", 2);
				SendMessage(wsci, SCI_STYLESETFORE, 2, RGB(0xff,0xff,0xff));
				SendMessage(wsci, SCI_STYLESETBACK, 2, RGB(0,0,0x80));
				SendMessage(wsci, SCI_STYLESETSIZE, 2, 12);
				SendMessage(wsci, SCI_STYLESETITALIC, 2, 1);
				AddStyledText(wsci, "December 1998 - April 1999.\n", 3);
				SendMessage(wsci, SCI_STYLESETFORE, 3, RGB(0xff,0xff,0xff));
				SendMessage(wsci, SCI_STYLESETBACK, 3, RGB(0,0,0x80));
				SendMessage(wsci, SCI_STYLESETSIZE, 3, 12);
				AddStyledText(wsci, "http://hare.net.au/~neilh/ScintillaTide.html", 4);
				SendMessage(wsci, SCI_STYLESETFORE, 4, RGB(0,0xff,0xff));
				SendMessage(wsci, SCI_STYLESETBACK, 4, RGB(0,0,0x80));
				SendMessage(wsci, SCI_STYLESETSIZE, 4, 9);
			}
		}
		return TRUE;

	case WM_CLOSE:
		SendMessage(hDlg,WM_COMMAND,IDCANCEL,NULL);
		break;

	case WM_COMMAND:
		if (LOWORD(wParam) == IDOK) {
			EndDialog(hDlg,IDOK);
			return TRUE;
		} else if (LOWORD(wParam) == IDCANCEL) {
			EndDialog(hDlg,IDCANCEL);
			return FALSE;
		}
	}

	return FALSE;
}

int DoDialog(HANDLE hInst,LPSTR lpResName,HWND hWnd,FARPROC lpProc,
             DWORD dwInitParam) {
	int result = -1;

	if (lpProc == NULL)
		lpProc = (FARPROC)DefaultDlg;

	DLGPROC lpFn = (DLGPROC)MakeProcInstance(lpProc,hInst);
	if (lpFn == NULL) {
		MessageBox(hWnd,"Failed to create proc instance.","Tide",MB_OK);
	} else {
		if (!dwInitParam)
			result = DialogBox((HINSTANCE)hInst,lpResName,hWnd,lpFn);
		else
			result = DialogBoxParam((HINSTANCE)hInst,lpResName,hWnd,lpFn,dwInitParam);

		if (result == -1) {
			DWORD dwError = GetLastError();
			char errormsg[200];
			sprintf(errormsg, "Failed to create Dialog box %d.", dwError);
			MessageBox(hWnd, errormsg, "Tide", MB_OK);
		}

		FreeProcInstance(lpFn);
	}

	return result;
}
#endif

//#define TRACE

static void dprintf(char *format, ...) {
#ifdef TRACE
	char buffer[2000];
	char *pArguments = (char *) & format + sizeof(format);
	vsprintf(buffer,format,pArguments);
#ifdef GTK
	printf("%s",buffer);
#else
	OutputDebugString(buffer);
#endif
#endif
}

const char propGlobalFileName[]="TideGlobal.properties";
const char propFileName[]="Tide.properties";

class RecentFile {
public:
	char fileName[MAX_PATH];
	char menuPath[MAX_PATH + 10];
	int lineNumber;
	RecentFile() {
		fileName[0] = '\0';
		menuPath[0] = '\0';
		lineNumber = -1;
	}
};

class TideWindow {
#ifndef GTK
	HINSTANCE hInstance;
	LPSTR cmdLine;
	static char *className;
#endif
	char windowName[MAX_PATH + 20];
	char fullPath[MAX_PATH];
	char fileName[MAX_PATH];
	char fileExt[MAX_PATH];
	char dirName[MAX_PATH];
	char dirNameAtExecute[MAX_PATH];

	enum { fileStackMax = 10 };
	RecentFile recentFileStack[fileStackMax];
	enum { fileStackCmdID = 2 };

	char findWhat[200];
	char replaceWhat[200];
#ifndef GTK
	FINDREPLACE fr;
#endif
	HWND wFindReplace;
	bool replacing;
	bool havefound;
	bool matchCase;
	bool wholeWord;

	char *language;
	char **keywords;
	char *keywordlist;
	char *functionDefinition;
#ifndef GTK
	HMENU pyroMenu;
#endif
	int currentStep;

	HWND hwndTide;
	HWND hwndEditor;
	HWND hwndOutput;

	int heightOutput;
	int heightOutputStartDrag;
	POINT ptStartDrag;
	bool capturedMouse;
	bool firstPropertiesRead;
	bool splitVertical;
	bool selMargin;
	bool bufferedDraw;

	bool isDirty;
	bool isBuilt;
	bool executing;
	enum { commandMax = 2 };
	int commandCurrent;
	char *commandToExecute[commandMax];
	long cancelFlag;

#ifdef GTK
	// Control of sub process
	int icmd;
	int originalEnd;
	int fdFIFO;
	int pidShell;
	char resultsFile[200];
	int inputHandle;
#endif

	PropSet propsBase;
	PropSet props;

	enum { heightBar = 7 };

#ifdef GTK
	GtkWidget *window;
	GtkWidget *fileSelector;
	GtkWidget *findInFilesDialog;
	GtkWidget *entryFiles;
	GtkWidget *gotoDialog;
	GtkWidget *gotoEntry;
	GtkWidget *toggleWord;
	GtkWidget *toggleCase;
	GtkWidget *entryFind;
	GtkWidget *entryReplace;
	GtkItemFactory *itemFactory;
	GtkAccelGroup *accelGroup;
#endif

	void ReadGlobalPropFile();
	void ReadLocalPropFile();

	LRESULT SendEditor(UINT msg, WPARAM wParam=0, LPARAM lParam=0);
	LRESULT SendOutput(UINT msg, WPARAM wParam=0, LPARAM lParam=0);
	void SendChildren(UINT msg, WPARAM wParam=0, LPARAM lParam=0);
	int LengthDocument();
	void GetRange(HWND win,int start, int end, char *text);
	void Colourise(int start=0, int end=-1, bool editor=true);

	void SetWindowName();
	void SetFileName(char *openName);
	void New();
	bool Exists(const char *dir, const char *path, char *testPath);
	void Open(char *file=0,bool initialCmdLine=false);
	int SaveIfUnsure();
	void Save();
	void SaveAs(char *file=0);
	void OpenProperties(bool local);
	void HandleFindReplace();
	void SelectionIntoFind();
	void Find();
	void FindNext();
	void FindInFiles();
	void Replace();
	void ReplaceOnce();
	void ReplaceAll();
	void GoLineDialog();
#ifdef GTK
	void FindReplaceGrabFields();
#endif
	void FindReplace(bool replace);
	void OutputAppendString(char *s, int len = -1);
#ifdef GTK
	void ContinueExecute();
#endif
	void Execute();
	void GoMessage(int dir);
	void StartCallTip();
	void ContinueCallTip();
	void StartAutoComplete();
	void SourceChanged();
	void CharAdded(char ch);
	int GetCurrentLineNumber();
	void AddCommand(char *cmd);
	void Command(WPARAM wParam, LPARAM lParam);
	void Notify(SCNotification *notification);

	void DoPop(HWND hWnd,POINT p);

	void CheckMenus();
	void SizeSubWindows();

	void DeleteFileStackMenu();
	void SetFileStackMenu();
	void DropFileStackTop();
	void AddFileToStack(char *file, int line=-1);
	void StackMenu(int pos);

	void ReadProperties();
	void SetStyleFor(HWND win, char *language);

	void Redraw();
	int NormaliseSplit(int splitPos);
	void MoveSplit(POINT ptNewDrag);

#ifdef GTK
	// GTK+ Signal Handlers

	static void OpenCancelSignal(GtkWidget *w, TideWindow *tidew);
	static void OpenOKSignal(GtkWidget *w, TideWindow *tidew);
	static void SaveAsSignal(GtkWidget *w, TideWindow *tidew);

	static void FindInFilesSignal(GtkWidget *w, TideWindow *tidew);
	static void FindInFilesCancelSignal(GtkWidget *w, TideWindow *tidew);

	static void GotoCancelSignal(GtkWidget *w, TideWindow *tidew);
	static void GotoSignal(GtkWidget *w, TideWindow *tidew);

	static void FRCancelSignal(GtkWidget *w, TideWindow *tidew);
	static void FRFindSignal(GtkWidget *w, TideWindow *tidew);
	static void FRReplaceSignal(GtkWidget *w, TideWindow *tidew);
	static void FRReplaceAllSignal(GtkWidget *w, TideWindow *tidew);

	static void IOSignal(TideWindow *tidew);

	static void QuitSignal(TideWindow *tidew, guint action, GtkWidget *w);
	static void MenuSignal(TideWindow *tidew, guint action, GtkWidget *w);
	static void CommandSignal(GtkWidget *w, gint wParam, gpointer lParam, TideWindow *tidew);
	static void NotifySignal(GtkWidget *w, gint wParam, gpointer lParam, TideWindow *tidew);
#endif

public:

#ifdef GTK
	TideWindow();
#else
	TideWindow(HINSTANCE hInstance_, LPSTR cmdLine_);
#endif
	~TideWindow();

	long WndProc(WORD iMessage,WPARAM wParam,LPARAM lParam);
#ifdef GTK
	void Run(char *cmdLine);
#else
	static void Register(HINSTANCE hInstance);
	static LRESULT PASCAL TWndProc(
		HWND hWnd,UINT iMessage,WPARAM wParam, LPARAM lParam);
	void ProcessExecute();
	HWND GetWnd();
#endif
};

#ifndef GTK
char *TideWindow::className=NULL;
#endif

#ifdef GTK
TideWindow::TideWindow() {
#else
	TideWindow::TideWindow(HINSTANCE hInstance_, LPSTR cmdLine_) {
#endif
		//dprintf("Constructing\n");
#ifndef GTK
		hInstance = hInstance_;
		cmdLine = cmdLine_;
		pyroMenu = 0;
#endif

		language = strdup("java");
		keywords = 0;
		keywordlist = 0;
		functionDefinition = 0;
		currentStep = 0;

		heightOutput = 0;

		isDirty = false;
		isBuilt = false;
		executing = false;
		commandCurrent = 0;
		for (int ic=0; ic<commandMax; ic++)
			commandToExecute[ic] = 0;
		InterlockedExchange(&cancelFlag, 0L);

#ifdef GTK
		// Control of sub process
		icmd = 0;
		originalEnd = 0;
		fdFIFO = 0;
		pidShell = 0;
		sprintf(resultsFile, "/tmp/Tide%x.results",
        		static_cast<int>(getpid()));
		inputHandle = 0;
#endif

		ptStartDrag.x = 0;
		ptStartDrag.y = 0;
		capturedMouse = false;
		firstPropertiesRead = true;
		splitVertical = false;
		selMargin = true;
		bufferedDraw = true;

		findWhat[0] = '\0';
		replaceWhat[0] = '\0';
#ifndef GTK
		memset(&fr, 0, sizeof(fr));
#endif
		wFindReplace = NULL;
		replacing = false;
		havefound = false;
		matchCase = false;
		wholeWord = false;

		windowName[0] = '\0';
		fullPath[0] = '\0';
		fileName[0] = '\0';
		fileExt[0] = '\0';
		dirName[0] = '\0';

		props.superPS = &propsBase;

		hwndTide = 0;
		hwndEditor = 0;
		hwndOutput = 0;

		ReadGlobalPropFile();

#ifdef GTK
		window = 0;
		fileSelector = 0;
		wFindReplace = 0;
		findInFilesDialog = 0;
		entryFiles = 0;
		gotoDialog = 0;
		gotoEntry = 0;
		toggleWord = 0;
		toggleCase = 0;
		entryFind = 0;
		entryReplace = 0;
		itemFactory = 0;
#else
		// Pass 'this' pointer in lpParam of CreateWindow().
		int left = props.GetInt("position.left");
		int top = props.GetInt("position.top");
		int width = props.GetInt("position.width");
		int height = props.GetInt("position.height");
		hwndTide = CreateWindowEx(
               		WS_EX_CLIENTEDGE,
               		className,
               		windowName,
               		WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
               		WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
               		WS_MAXIMIZE | WS_CLIPCHILDREN,
               		left ? left : CW_USEDEFAULT,
               		top ? top : CW_USEDEFAULT,
               		width ? width : CW_USEDEFAULT,
               		height ? height : CW_USEDEFAULT,
               		NULL,
               		NULL,
               		hInstance,
               		(LPSTR) this);
		if (!hwndTide)
			exit(FALSE);

		Open(cmdLine,true);

		ShowWindow(hwndTide, SW_SHOWNORMAL);
#endif
	}

#ifndef GTK
	void TideWindow::Register(HINSTANCE hInstance) {
		const char *AppName="Tide";

		WNDCLASS wndclass;   // Structure used to register Windows class.

		className = "TideWindow";

		wndclass.style         = CS_HREDRAW | CS_VREDRAW;
		wndclass.lpfnWndProc   = TideWindow::TWndProc;
		wndclass.cbClsExtra    = 0;
		wndclass.cbWndExtra    = sizeof(TideWindow*);
		wndclass.hInstance     = hInstance;
		wndclass.hIcon         = LoadIcon(hInstance,(LPSTR)AppName);
		wndclass.hCursor       =  NULL;
		wndclass.hbrBackground = NULL;
		wndclass.lpszMenuName  = (LPSTR)AppName;
		wndclass.lpszClassName = className;

		if (!RegisterClass(&wndclass))
			exit(FALSE);
	}

#endif

	bool GetDefaultPropertiesFileName(char *pathDefaultProps, int lenPath) {
#ifdef GTK
		strcpy(pathDefaultProps, getenv("HOME"));
		strcat(pathDefaultProps, pathSepString);
		strcat(pathDefaultProps, propGlobalFileName);
		return true;
#else
		GetModuleFileName(0, pathDefaultProps, lenPath);
		char *lastSlash = strrchr(pathDefaultProps, pathSepChar);
		if (lastSlash && ((lastSlash +1 - pathDefaultProps + strlen(propGlobalFileName)) < lenPath)) {
			strcpy(lastSlash + 1, propGlobalFileName);
			return true;
		} else {
			return false;
		}
#endif
	}

	void TideWindow::ReadGlobalPropFile() {
		char propfile[MAX_PATH+20];
		if (GetDefaultPropertiesFileName(propfile, sizeof(propfile))) {
			propsBase.Read(propfile);
		}
	}

	void TideWindow::ReadLocalPropFile() {
		char propfile[MAX_PATH+20];
		if (dirName[0])
			strcpy(propfile, dirName);
		else
			getcwd(propfile, sizeof(propfile));
		strcat(propfile, pathSepString);
		strcat(propfile, propFileName);
		props.Read(propfile);
		dprintf("Reading local properties from %s\n", propfile);
	}

#ifdef GTK
	LRESULT SendMessage(GtkWidget *w,UINT msg, WPARAM wParam=0, LPARAM lParam=0) {
		return scintilla_send_message(SCINTILLA(w), msg, wParam, lParam);
	}

	void SetFocus(GtkWidget *hwnd) {
		SendMessage(hwnd, SCI_GRABFOCUS, 0, 0);
	}

	void DrawMenuBar(GtkWidget *hwnd) {
	}
#endif

	LRESULT TideWindow::SendEditor(UINT msg, WPARAM wParam, LPARAM lParam) {
		return SendMessage(hwndEditor, msg, wParam, lParam);
	}

	LRESULT TideWindow::SendOutput(UINT msg, WPARAM wParam, LPARAM lParam) {
		return SendMessage(hwndOutput, msg, wParam, lParam);
	}

	void TideWindow::SendChildren(UINT msg, WPARAM wParam, LPARAM lParam) {
		SendEditor(msg, wParam, lParam);
		SendOutput(msg, wParam, lParam);
	}

	void TideWindow::DeleteFileStackMenu() {
#ifdef GTK
		for (int stackPos=1;stackPos<fileStackMax;stackPos++) {
			if (recentFileStack[stackPos].menuPath[0]) {
				gtk_item_factory_delete_item(itemFactory,
                             				recentFileStack[stackPos].menuPath);
				recentFileStack[stackPos].menuPath[0] = '\0';
			}
		}
		gtk_item_factory_delete_item(itemFactory, "/File/sep2");
		gtk_item_factory_delete_item(itemFactory, "/File/Quit");
#endif
	}

	static void strSafeMenuItem(char *dest, const char *source) {
		dest += strlen(dest);
		while (*source) {
			if (*source == '_')
				*dest++ = '_';
			*dest++ = *source++;
		}
		*dest++ = '\0';
	}

	void TideWindow::SetFileStackMenu() {
#ifdef GTK
		GtkItemFactoryCallback menuSig = GtkItemFactoryCallback(MenuSignal);
		for (int stackPos=1;stackPos<fileStackMax;stackPos++) {
			int itemID=fileStackCmdID+stackPos;
			if (recentFileStack[stackPos].fileName[0]) {
				strcpy(recentFileStack[stackPos].menuPath, "/File/");
				char *cpDirEnd = strrchr(recentFileStack[stackPos].fileName, pathSepChar);
				if (cpDirEnd) {
					strSafeMenuItem(recentFileStack[stackPos].menuPath, cpDirEnd+1);
				} else {
					strSafeMenuItem(recentFileStack[stackPos].menuPath, recentFileStack[stackPos].fileName);
				}
				GtkItemFactoryEntry itemFile = {
    				recentFileStack[stackPos].menuPath,
    				NULL,
    				menuSig,
    				itemID,
    				NULL
				};
				gtk_item_factory_create_item(itemFactory, &itemFile, this, 1);
			}
		}
		if (recentFileStack[1].fileName[0]) {
			GtkItemFactoryEntry itemSep = {"/File/sep2", NULL, NULL, 0, "<Separator>"
                              			};
			gtk_item_factory_create_item(itemFactory, &itemSep, "", 1);
		}
		GtkItemFactoryEntry itemQuit = {"/File/_Quit", "<control>Q", menuSig, IDM_QUIT, 0
                               		};
		gtk_item_factory_create_item(itemFactory, &itemQuit, this, 1);
#else
		const int startStackMenuPos = 5;
		HMENU hmenu = GetMenu(hwndTide);
		HMENU hmenuFile = GetSubMenu(hmenu,0);
		// If there is no file stack menu item and there are > 1 files in stack
		if ((GetMenuState(hmenuFile, fileStackCmdID, MF_BYCOMMAND) == 0xffffffff) &&
    			(recentFileStack[1].fileName[0] != '\0')) {
			InsertMenu(hmenuFile, startStackMenuPos, MF_BYPOSITION|MF_SEPARATOR, fileStackCmdID,"-");
		}
		for (int stackPos=1;stackPos<fileStackMax;stackPos++) {
			int itemID=fileStackCmdID+stackPos;
			if (recentFileStack[stackPos].fileName[0]) {
				//dprintf("FileStack[%d] = <%s>\n", stackPos, fileStack[stackPos]);
				if (GetMenuState(hmenuFile, itemID, MF_BYCOMMAND) == 0xffffffff) {
					InsertMenu(hmenuFile, startStackMenuPos + stackPos, MF_BYPOSITION, itemID,
           					recentFileStack[stackPos].fileName);
				} else {
					ModifyMenu(hmenuFile, itemID, MF_BYCOMMAND, itemID,
           					recentFileStack[stackPos].fileName);
				}
			} else {
				if (GetMenuState(hmenuFile, itemID, MF_BYCOMMAND) != 0xffffffff) {
					DeleteMenu(hmenuFile,itemID, MF_BYCOMMAND);
				}
			}
		}
#endif
	}

	void TideWindow::DropFileStackTop() {
		DeleteFileStackMenu();
		for (int stackPos=0;stackPos<fileStackMax;stackPos++)
			recentFileStack[stackPos] = recentFileStack[stackPos+1];
		strcpy(recentFileStack[fileStackMax-1].fileName,"");
		recentFileStack[fileStackMax-1].lineNumber = -1;
		SetFileStackMenu();
	}

	void TideWindow::AddFileToStack(char *file, int line) {
		DeleteFileStackMenu();
		if (file[strlen(file) - 1] == '\\')
			return;
		int stackPos;
		int eqPos=fileStackMax-1;
		for (stackPos=0;stackPos<fileStackMax;stackPos++)
			if (strcmp(recentFileStack[stackPos].fileName,file) == 0)
				eqPos = stackPos;
		for (stackPos=eqPos;stackPos>0;stackPos--)
			recentFileStack[stackPos] = recentFileStack[stackPos-1];
		strcpy(recentFileStack[0].fileName,file);
		recentFileStack[0].lineNumber = line;
		SetFileStackMenu();
		DrawMenuBar(hwndTide);
	}

	void TideWindow::StackMenu(int pos) {
		//dprintf("Stack menu\n");
		if (pos == -1) {
			int lastPos = -1;
			for (int stackPos=1;stackPos<fileStackMax;stackPos++) {
				if (recentFileStack[stackPos].fileName[0])
					lastPos = stackPos;
			}
			if (lastPos > 0) {
				//dprintf("Opening lpos %d %s\n",recentFileStack[lastPos].lineNumber,recentFileStack[lastPos].fileName);
				int line = recentFileStack[lastPos].lineNumber;
				Open(recentFileStack[lastPos].fileName);
				if (line != -1)
					SendEditor(SCI_GOTOLINE, line + 1);
			}
		} else {
			if (recentFileStack[pos].fileName[0] == '\0')	// Empty
				New();
			else {
				int line = recentFileStack[pos].lineNumber;
				//dprintf("Opening pos %d %s\n",recentFileStack[pos].lineNumber,recentFileStack[pos].fileName);
				Open(recentFileStack[pos].fileName);
				if (line != -1)
					SendEditor(SCI_GOTOLINE, line + 1);
			}
		}
	}

	static int IntFromHexDigit(const char ch) {
		if (isdigit(ch))
			return ch - '0';
		else if (ch >= 'A' && ch <= 'F')
			return ch - 'A' + 10;
		else if (ch >= 'a' && ch <= 'f')
			return ch - 'a' + 10;
		else
			return 0;
	}

	static long ColourFromString(const char *val) {
		int r = IntFromHexDigit(val[1]) * 16 + IntFromHexDigit(val[2]);
		int g = IntFromHexDigit(val[3]) * 16 + IntFromHexDigit(val[4]);
		int b = IntFromHexDigit(val[5]) * 16 + IntFromHexDigit(val[6]);
		return LRGB(r,g,b);
	}

	void TideWindow::SetStyleFor(HWND win, char *language) {
		SendMessage(win, SCI_STYLECLEARALL, 0, 0);

		for (int style=0;style < 32; style++) {
			char *val = "";
			char key[200];
			sprintf(key, "style.%s.%0d", language, style);
			val = strdup(props.Get(key));
			char *opt = val;
			while (opt) {
				char *cpComma = strchr(opt,',');
				if (cpComma)
					*cpComma = '\0';
				char *colon = strchr(opt, ':');
				if (colon)
					*colon++ = '\0';
				if (0 == strcmp(opt, "italics"))
					SendMessage(win, SCI_STYLESETITALIC, style, 1);
				if (0 == strcmp(opt, "bold"))
					SendMessage(win, SCI_STYLESETBOLD, style, 1);
				if (0 == strcmp(opt, "font"))
					SendMessage(win, SCI_STYLESETFONT, style, reinterpret_cast<LPARAM>(colon));
				if (0 == strcmp(opt, "fore"))
					SendMessage(win, SCI_STYLESETFORE, style, ColourFromString(colon));
				if (0 == strcmp(opt, "back"))
					SendMessage(win, SCI_STYLESETBACK, style, ColourFromString(colon));
				if (0 == strcmp(opt, "size"))
					SendMessage(win, SCI_STYLESETSIZE, style, atoi(colon));
				if (cpComma)
					opt = cpComma + 1;
				else
					opt = 0;
			}
			if (val)
				free(val);
		}
	}

	void TideWindow::ReadProperties() {
		DWORD dwStart = timeGetTime();
		free(language);
		language = props.GetNewExpand("lexer.", fileName);
		if (keywords) {
			delete []keywords;
			free(keywordlist);
		}
		keywordlist = props.GetNewExpand("keywords.", fileName);
		char prev = ' ';
		int words = 0;
		for (int j=0; keywordlist[j]; j++) {
			if (!isspace(keywordlist[j]) && isspace(prev))
				words++;
			prev = keywordlist[j];
		}
		keywords = new char*[words+1];
		words = 0;
		prev = '\0';
		int len = strlen(keywordlist);
		for (int k=0; k < len; k++) {
			if (!isspace(keywordlist[k])) {
				if (!prev) {
					keywords[words] = &keywordlist[k];
					words++;
				}
			} else {
				keywordlist[k] = '\0';
			}
			prev = keywordlist[k];
		}
		keywords[words] = &keywordlist[len];

		int codePage = props.GetInt("code.page");
		SendEditor(SCI_SETCODEPAGE, codePage);

		char *colour = 0;

		colour = props.Get("back");
		if (colour && *colour) {
			SendEditor(SCI_SETBACK, ColourFromString(colour));
		}
		colour = props.Get("fore");
		if (colour && *colour) {
			SendEditor(SCI_SETFORE, ColourFromString(colour));
		}

		char *size = props.Get("size");
		if (size && *size) {
			SendChildren(SCI_SETSIZE, atoi(size));
		}

		char *font = props.Get("font");
		if (size && *size) {
			SendChildren(SCI_SETFONT, reinterpret_cast<WPARAM>(font));
		}

		char *selfore = props.Get("selection.fore");
		if (selfore && *selfore) {
			SendChildren(SCI_SETSELFORE, 1, ColourFromString(selfore));
		} else {
			SendChildren(SCI_SETSELFORE, 0, 0);
		}
		colour = props.Get("selection.back");
		if (colour && *colour) {
			SendChildren(SCI_SETSELBACK, 1, ColourFromString(colour));
		} else {
			if (selfore && *selfore)
				SendChildren(SCI_SETSELBACK, 0, 0);
			else	// Have to show selection somehow
				SendChildren(SCI_SETSELBACK, 1, LRGB(0xC0,0xC0,0xC0));
		}


		SetStyleFor(hwndEditor, language);
		SetStyleFor(hwndOutput, "errorlist");

		// As user can change vertical versus horizontal split, only read at app startup
		if (firstPropertiesRead) {
			splitVertical = props.GetInt("split.vertical");
			SendEditor(SCI_SETVIEWWS, props.GetInt("view.whitespace"));
		}

#ifndef GTK
		HMENU mainm = GetMenu(hwndTide);
#endif
		if (0 == strcmp(language, "pyro")) {
			SendEditor(SCI_MARKERADD, 7, 0);
#ifndef GTK
			if (0 == pyroMenu) {
				pyroMenu = CreatePopupMenu();
				InsertMenu(mainm, 3, MF_POPUP|MF_BYPOSITION, (int)pyroMenu, "&Pyro");
				AppendMenu(pyroMenu, MF_STRING, IDM_TABTIMMY, "TabTimmy");
				AppendMenu(pyroMenu, MF_STRING, IDM_SELECTIONMARGIN, "Selection Margin");
				AppendMenu(pyroMenu, MF_STRING, IDM_BUFFEREDDRAW, "Buffered Drawing");
				AppendMenu(pyroMenu, MF_STRING, IDM_STEP, "Step\tF8");
			}
#endif
		} else {
			SendEditor(SCI_MARKERDELETEALL, -1);
#ifndef GTK
			// Get rid of menu
			if (pyroMenu) {
				RemoveMenu(mainm, 3, MF_BYPOSITION);
				DestroyMenu(pyroMenu);
				pyroMenu = 0;
			}
#endif
		}
		int tabSize = props.GetInt("tabsize");
		if (tabSize) {
			SendEditor(SCI_SETTABWIDTH, tabSize);
		}

		firstPropertiesRead = false;

		DWORD dwEnd = timeGetTime();
		//dprintf("end language find %d\n", dwEnd - dwStart);
		//dprintf("Language find for <%s> is <%s>\n", fileExt, language);
	}

#ifdef GTK
	int GetWindowTextLength(HWND w) {
		return SendMessage(w, WM_GETTEXTLENGTH);
	}
#endif

	int TideWindow::LengthDocument() {
		return SendEditor(SCI_GETLENGTH);
	}

	void TideWindow::GetRange(HWND win,int start, int end, char *text) {
		TEXTRANGE tr;
		tr.chrg.cpMin = start;
		tr.chrg.cpMax = end;
		tr.lpstrText = text;
		SendMessage(win, EM_GETTEXTRANGE, 0, reinterpret_cast<LPARAM>(&tr));
	}

	void TideWindow::Colourise(int start, int end, bool editor) {
		DWORD dwStart = timeGetTime();
		HWND win = editor ? hwndEditor : hwndOutput;
		int lengthDoc = SendMessage(win, SCI_GETLENGTH,0,0);
		if (end == -1)
			end = lengthDoc;
		char *cdoc = new char[end - start + 1];
		//dprintf("Colourise size=%d ptr=%x lan=%s\n", lengthDoc, cdoc, language);

		// Have to use SendMessage rather than GetWindowText as GetWindowText truncates
		// length parameter to 16 bits
		GetRange(win, start, end, cdoc);

		int styleStart = 0;
		if (start > 0)
			styleStart = SendMessage(win, SCI_GETSTYLEAT, start - 1, 0);
		if (editor)
			ColouriseDoc(cdoc,start,end-start,styleStart,language,keywords,win);
		else
			ColouriseDoc(cdoc,start,end-start,0,"errorlist",0,win);
		Redraw();
		delete []cdoc;
		DWORD dwEnd = timeGetTime();
		//dprintf("end colourise %d\n", dwEnd - dwStart);
	}

	TideWindow::~TideWindow() {
		free(language);
		language = 0;
		free(keywords);
		keywords = 0;
		free(keywordlist);
		keywordlist = 0;
		free(functionDefinition);
		functionDefinition = 0;
	};

	void TideWindow::SetWindowName() {
		if (fileName[0] == '\0')
			strcpy(windowName,"(Untitled)");
		else
			strcpy(windowName,fileName);
		strcat(windowName," - Tide");
#ifdef GTK
		gtk_window_set_title(GTK_WINDOW(window), windowName);
#else
		SetWindowText(hwndTide,windowName);
#endif
	}

#ifdef GTK
	static void _fullpath(char *fullPath, const char *basePath, int size) {
		strncpy(fullPath, basePath, size);
	}
#endif

	void TideWindow::SetFileName(char *openName) {
		char pathCopy[MAX_PATH+1];
		pathCopy[0] = '\0';

		if (openName[0] == '\"') {
			strncpy(pathCopy, openName+1,MAX_PATH);
			pathCopy[MAX_PATH] = '\0';
			if (pathCopy[strlen(pathCopy) - 1] == '\"')
				pathCopy[strlen(pathCopy) - 1] = '\0';
			_fullpath(fullPath, pathCopy, MAX_PATH);
		} else if (openName[0]) {
			_fullpath(fullPath, openName, MAX_PATH);
		} else {
			fullPath[0] = '\0';
		}

		char *cpDirEnd = strrchr(fullPath, pathSepChar);
		if (cpDirEnd) {
			strcpy(fileName, cpDirEnd+1);
			strcpy(dirName, fullPath);
			dirName[cpDirEnd - fullPath] = '\0';
		} else {
			strcpy(fileName, fullPath);
			getcwd(dirName, sizeof(dirName));
			//dprintf("Working directory: <%s>\n", dirName);
			strcpy(fullPath, dirName);
			strcat(fullPath, pathSepString);
			strcat(fullPath, fileName);
		}
		char fileBase[MAX_PATH];
		strcpy(fileBase, fileName);
		char *cpExt = strrchr(fileBase, '.');
		if (cpExt) {
			*cpExt = '\0';
			strcpy(fileExt, cpExt + 1);
		} else { // No extension
			fileExt[0] = '\0';
		}

		ReadLocalPropFile();

		props.Set("FilePath", fullPath);
		props.Set("FileDir", dirName);
		props.Set("FileName", fileBase);
		props.Set("FileExt", fileExt);
		props.Set("FileNameExt", fileName);

		SetWindowName();
	}

#ifdef GTK
	void TideWindow::OpenCancelSignal(GtkWidget *w, TideWindow *tidew) {
		gtk_widget_destroy(GTK_WIDGET(tidew->fileSelector));
		tidew->fileSelector = 0;
	}

	void TideWindow::OpenOKSignal(GtkWidget *w, TideWindow *tidew) {
		tidew->Open(gtk_file_selection_get_filename(
                		GTK_FILE_SELECTION(tidew->fileSelector)));
		gtk_widget_destroy(GTK_WIDGET(tidew->fileSelector));
		tidew->fileSelector = 0;
	}

	void TideWindow::SaveAsSignal(GtkWidget *w, TideWindow *tidew) {
		printf("Do Save AS\n");
		tidew->SaveAs(gtk_file_selection_get_filename(
                  		GTK_FILE_SELECTION(tidew->fileSelector)));
	}
#endif

	void TideWindow::New() {
		SendEditor(SCI_CLEARALL);
		fullPath[0] = '\0';
		fileName[0] = '\0';
		fileExt[0] = '\0';
		dirName[0] = '\0';
		SetWindowName();
		isDirty = false;
		isBuilt = false;

		// Define markers
		//#ifdef WANT_MARKER_DEMO
		SendEditor(SCI_MARKERDEFINE, 0, SC_MARK_CIRCLE);
		SendEditor(SCI_MARKERSETFORE, 0, LRGB(0x7f, 0, 0));
		SendEditor(SCI_MARKERSETBACK, 0, LRGB(0xff, 0, 0));

		SendEditor(SCI_MARKERDEFINE, 1, SC_MARK_ROUNDRECT);
		SendEditor(SCI_MARKERSETFORE, 1, LRGB(0xff, 0xff, 0));
		SendEditor(SCI_MARKERSETBACK, 1, LRGB(0, 0, 0x7f));

		SendEditor(SCI_MARKERDEFINE, 2, SC_MARK_SHORTARROW);
		SendEditor(SCI_MARKERSETFORE, 2, LRGB(0, 0x7f, 0x7f));
		SendEditor(SCI_MARKERSETBACK, 2, LRGB(0, 0xff, 0));
		//#endif
	}

	bool TideWindow::Exists(const char *dir, const char *path, char *testPath) {
		if (dir) {
			strcpy(testPath, path);
		} else {
			if ((strlen(dir) + strlen(pathSepString) + strlen(path) + 1) > MAX_PATH)
				return false;
			strcpy(testPath, dir);
			strcat(testPath, pathSepString);
			strcat(testPath, path);
		}
		FILE *fp = fopen(testPath,"rb");
		if (!fp)
			return false;
		fclose(fp);
		return true;
	}

	const int blockSize = 32768;

	void TideWindow::Open(char *file,bool initialCmdLine) {
		if (file) {
			New();
			//dprintf("Opening %s\n", file);
			SetFileName(file);
			ReadProperties();
			if (fileName[0]) {
				DWORD dwStart = timeGetTime();
				SendEditor(SCI_SETUNDOCOLLECTION, 0);
				FILE *fp = fopen(fullPath,"rb");
				if (fp || initialCmdLine) {
					if (fp) {
						char data[blockSize];
						char styledData[blockSize*2];
						memset(styledData, 0, sizeof(styledData));
						AddFileToStack(fullPath);
						int lenFile = fread(data,1,sizeof(data),fp);
						int curEnd = 0;
						while (lenFile > 0) {
							//dprintf("opening %s %d\n", fullPath, lenFile);
							for (int i = 0; i < lenFile; i++)
								styledData[i*2] = data[i];
							SendEditor(SCI_ADDSTYLEDTEXT, lenFile*2,
           							(LPARAM)styledData);
							curEnd += lenFile*2;
							lenFile = fread(data,1,sizeof(data),fp);
						}
						fclose(fp);
						isDirty = false;
					} else {
						isDirty = true;
					}
					isBuilt = false;
					DWORD dwEnd = timeGetTime();
					//dprintf("end read %d\n", dwEnd - dwStart);

					SendEditor(SCI_SETUNDOCOLLECTION, 1);

					if (0 == strcmp(language, "pyro")) {
						SendEditor(SCI_MARKERADD, 7, 0);
					}
				} else {
					char msg[200];
					strcpy(msg, "Could not open file \"");
					strcat(msg, fullPath);
					strcat(msg, "\".");
					MessageBox(hwndTide,msg,"Tide",MB_OK);
				}
				SendEditor(EM_EMPTYUNDOBUFFER);
				SendEditor(SCI_GOTOPOS, 0);
			}
			Redraw();
		} else {
#ifdef GTK
			if (!fileSelector) {
				fileSelector = gtk_file_selection_new("Open File");
				gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->ok_button),
                   				"clicked", GtkSignalFunc(OpenOKSignal), this);
				gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->cancel_button),
                   				"clicked", GtkSignalFunc(OpenCancelSignal), this);
				// Other ways to destroy
				gtk_widget_show(fileSelector);
				gtk_grab_add(fileSelector);
			}
#else
			char openName[MAX_PATH] = "\0";
			OPENFILENAME ofn = {sizeof(OPENFILENAME)};
			ofn.hwndOwner = hwndTide;
			ofn.hInstance = hInstance;
			ofn.lpstrFile = openName;
			ofn.nMaxFile = sizeof(openName);
			char *filter = 0;
			char *openFilter = props.Get("open.filter");
			if (openFilter) {
				filter = strdup(openFilter);
				for (int fc=0;filter[fc];fc++)
					if (filter[fc] == '|')
						filter[fc] = '\0';
			}
			ofn.lpstrFilter = filter;
			ofn.lpstrFileTitle = "Open File";
			ofn.Flags = OFN_HIDEREADONLY;

			if (GetOpenFileName(&ofn)) {
				free(filter);
				dprintf("Open: <%s>\n", openName);
				Open(openName);
			} else {
				free(filter);
				return;
			}
#endif
		}
		AddFileToStack(fullPath);
		SetWindowName();
	}

	int TideWindow::SaveIfUnsure() {
		if (isDirty) {
			if (props.GetInt("are.you.sure")) {
				char msg[200];
				strcpy(msg, "Save changes to \"");
				strcat(msg, fullPath);
				strcat(msg, "\"?");
				int decision = MessageBox(hwndTide,msg,"Tide",MB_YESNOCANCEL);
				if (decision == IDYES) {
					Save();
				}
			} else {
				Save();
			}
		}
	}

	void TideWindow::Save() {
		if (fileName[0]) {
			dprintf("Saving <%s>\n", fileName);
			FILE *fp = fopen(fullPath, "wb");
			if (fp) {
				AddFileToStack(fullPath, GetCurrentLineNumber());
				int lengthDoc = LengthDocument();
				for (int i=0;i<lengthDoc;i++) {
					char ch = SendEditor(SCI_GETCHARAT, i);
					fputc(ch,fp);
				}
				fclose(fp);
				isDirty = false;
				if ((0 == stricmp(fileName, propFileName)) || (0 == stricmp(fileName, propGlobalFileName))) {
					ReadGlobalPropFile();
					ReadLocalPropFile();
					ReadProperties();
					Redraw();
				}
			} else {
				char msg[200];
				strcpy(msg, "Could not save file \"");
				strcat(msg, fullPath);
				strcat(msg, "\".");
				MessageBox(hwndTide,msg,"Tide",MB_OK);
			}
		} else {
			SaveAs();
		}
	}

	void TideWindow::SaveAs(char *file) {
		if (file && *file) {
			SetFileName(file);
			Save();
#ifdef GTK
			if (fileSelector) {
				gtk_widget_destroy(GTK_WIDGET(fileSelector));
				fileSelector = 0;
			}
#endif
			ReadProperties();
			Colourise();	// In case extension was changed
			Redraw();
			SetWindowName();
		} else {
#ifdef GTK
			if (!fileSelector) {
				fileSelector = gtk_file_selection_new("Save File As");
				gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->ok_button),
                   				"clicked", GtkSignalFunc(SaveAsSignal), this);
				gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector)->cancel_button),
                   				"clicked", GtkSignalFunc(OpenCancelSignal), this);
				// Other ways to destroy
				gtk_widget_show(fileSelector);
				gtk_grab_add(fileSelector);
			}
#else
			char openName[MAX_PATH] = "\0";
			strcpy(openName, fileName);
			OPENFILENAME ofn = {sizeof(ofn)};
			ofn.hwndOwner = hwndTide;
			ofn.hInstance = hInstance;
			ofn.lpstrFile = openName;
			ofn.nMaxFile = sizeof(openName);
			ofn.lpstrFileTitle = "Open File";
			ofn.Flags = OFN_HIDEREADONLY;

			if (GetSaveFileName(&ofn)) {
				dprintf("Save: <%s>\n", openName);
				SetFileName(openName);
				Save();
				ReadProperties();
				Colourise();	// In case extension was changed
				InvalidateRect(hwndEditor,(LPRECT)NULL,TRUE);
			}
#endif
		}
	}

	void TideWindow::OpenProperties(bool local) {
		if (SaveIfUnsure() != IDCANCEL) {
			char propfile[MAX_PATH+20];
			if (local) {
				getcwd(propfile, sizeof(propfile));
				strcat(propfile, pathSepString);
				strcat(propfile, propFileName);
				Open(propfile);
			} else {
				if (GetDefaultPropertiesFileName(propfile, sizeof(propfile))) {
					Open(propfile);
				}
			}
		}
	}

	void TideWindow::HandleFindReplace() {
#ifndef GTK
		if (fr.Flags & FR_DIALOGTERM) {
			wFindReplace = 0;
			currentDlg = 0;
			dprintf("Cancel FindReplace\n");
		} else if (fr.Flags & FR_FINDNEXT) {
			dprintf("FindNext <%s> %x\n", fr.lpstrFindWhat, fr.Flags);
			FindNext();
		} else if (fr.Flags & FR_REPLACE) {
			if (havefound) {
				SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(fr.lpstrReplaceWith));
				havefound = false;
				dprintf("Replace <%s> -> <%s>\n", fr.lpstrFindWhat, fr.lpstrReplaceWith);
			}
			FindNext();
		} else if (fr.Flags & FR_REPLACEALL) {
			FINDTEXTEX ft;
			ft.chrg.cpMin = 0;
			ft.chrg.cpMax = LengthDocument();
			ft.lpstrText = fr.lpstrFindWhat;
			ft.chrgText.cpMin = 0;
			ft.chrgText.cpMax = 0;
			int posFind = SendEditor(EM_FINDTEXTEX, fr.Flags & (FR_WHOLEWORD | FR_MATCHCASE),
                         			reinterpret_cast<LPARAM>(&ft));
			if (posFind != -1) {
				while (posFind != -1) {
					SendEditor(EM_SETSEL, ft.chrgText.cpMax + 1, ft.chrgText.cpMin);
					SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(fr.lpstrReplaceWith));
					ft.chrg.cpMin = posFind + strlen(fr.lpstrReplaceWith) + 1;
					ft.chrg.cpMax = LengthDocument();
					posFind = SendEditor(EM_FINDTEXTEX, fr.Flags & (FR_WHOLEWORD | FR_MATCHCASE),
                     					reinterpret_cast<LPARAM>(&ft));
				}
			} else {
				char msg[200];
				strcpy(msg, "No replacements because string \"");
				strcat(msg, fr.lpstrFindWhat);
				strcat(msg, "\" was not present.");
				MessageBox(wFindReplace,msg,"Tide",MB_OK|MB_ICONWARNING);
			}
			dprintf("ReplaceAll <%s> -> <%s>\n", fr.lpstrFindWhat, fr.lpstrReplaceWith);
		} else {
			dprintf("Find/replacemessage %x\n", fr.Flags);
		}
#endif
	}

	void TideWindow::SelectionIntoFind() {
		int selStart = 0;
		int selEnd = 0;
		int lengthDoc = LengthDocument();
		SendEditor(EM_GETSEL, (WPARAM)&selStart, (LPARAM)&selEnd);
		if (selStart == selEnd) {
			// Try and find a word at the caret
			if (isalnum(SendEditor(SCI_GETCHARAT, selStart))) {
				while ((selStart > 0) && (isalnum(SendEditor(SCI_GETCHARAT, selStart-1))))
					selStart--;
				while ((selEnd < lengthDoc-1) && (isalnum(SendEditor(SCI_GETCHARAT, selEnd+1))))
					selEnd++;
				if (selStart < selEnd)
					selEnd++;	// Because normal selections end one past
			}
		}
		if ((selStart < selEnd) && ((selEnd - selStart) < sizeof(findWhat))) {
			GetRange(hwndEditor, selStart, selEnd - 1, findWhat);
			findWhat[selEnd - selStart] = '\0';
		}
	}

	void TideWindow::Find() {
#ifdef GTK
		SelectionIntoFind();
		FindReplace(false);
#else
		if (wFindReplace)
			return;
		SelectionIntoFind();

		memset(&fr, 0, sizeof(fr));
		fr.lStructSize = sizeof(fr);
		fr.hwndOwner = hwndTide;
		fr.hInstance = hInstance;
		fr.Flags = FR_NOUPDOWN;
		fr.lpstrFindWhat = findWhat;
		fr.wFindWhatLen = sizeof(findWhat);
		wFindReplace = FindText(&fr);
		replacing = false;
		currentDlg = wFindReplace;
		dprintf("Find %x %x %x\n", wFindReplace, &fr, this);
#endif
	}

	void TideWindow::FindNext() {
		if (!findWhat[0]) {
			Find();
			return;
		}
		FINDTEXTEX ft;
		ft.chrg.cpMin = SendEditor(SCI_GETCURRENTPOS) + 1;
		ft.chrg.cpMax = LengthDocument();
		ft.lpstrText = findWhat;
		ft.chrgText.cpMin = 0;
		ft.chrgText.cpMax = 0;
		int flags = (wholeWord ? FR_WHOLEWORD : 0) | (matchCase ? FR_MATCHCASE : 0);
		int posFind = SendEditor(EM_FINDTEXTEX, flags, reinterpret_cast<LPARAM>(&ft));
		//dprintf("found at %d\n", posFind);
		if (posFind == -1) {
			havefound = false;
			char msg[200];
			strcpy(msg, "Cannot find the string \"");
			strcat(msg, findWhat);
			strcat(msg, "\".");
			if (wFindReplace)
				MessageBox(wFindReplace,msg,"Tide",MB_OK|MB_ICONWARNING);
			else
				MessageBox(hwndTide,msg,"Tide",MB_OK);
		} else {
			havefound = true;
			SendEditor(EM_SETSEL, ft.chrgText.cpMax + 1, ft.chrgText.cpMin);
			if (!replacing) {
#ifdef GTK
				if (wFindReplace) {
					gtk_widget_destroy(GTK_WIDGET(wFindReplace));
					wFindReplace = 0;
				}
#else
				if (wFindReplace) {
					EndDialog(wFindReplace,IDCANCEL);
					wFindReplace = 0;
					currentDlg = 0;
				}
#endif
			}
		}
	}

#ifdef GTK
	void TideWindow::FindInFilesSignal(GtkWidget *w, TideWindow *tidew) {
		char *findEntry = gtk_entry_get_text(GTK_ENTRY(tidew->entryFind));
		tidew->props.Set("find.what", findEntry);
		char *filesEntry = gtk_entry_get_text(GTK_ENTRY(tidew->entryFiles));
		tidew->props.Set("find.files", filesEntry);

		gtk_widget_destroy(GTK_WIDGET(tidew->findInFilesDialog));
		tidew->findInFilesDialog = 0;

		printf("Grepping for <%s> in <%s>\n",
       		tidew->props.Get("find.what"),
       		tidew->props.Get("find.files"));
		tidew->AddCommand(strdup(tidew->props.GetNewExpand("find.command", "")));
		if (tidew->commandCurrent > 0)
			tidew->Execute();
	}

	void TideWindow::FindInFilesCancelSignal(GtkWidget *w, TideWindow *tidew) {
		gtk_widget_destroy(GTK_WIDGET(tidew->findInFilesDialog));
		tidew->findInFilesDialog = 0;
	}
#else
	int PASCAL GrepDlg(HWND hDlg,WORD message,WPARAM wParam,LONG lParam) {
		static PropSet *props;

		switch (message) {

		case WM_INITDIALOG:
			props = (PropSet *)lParam;
			SetDlgItemText(hDlg, IDFINDWHAT, props->Get("find.what"));
			SetDlgItemText(hDlg, IDFILES, props->Get("find.files"));
			//SetDlgItemText(hDlg, IDDIRECTORY, props->Get("find.directory"));
			return TRUE;

		case WM_CLOSE:
			SendMessage(hDlg,WM_COMMAND,IDCANCEL,NULL);
			break;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg,IDCANCEL);
				return FALSE;
			} else if (LOWORD(wParam) == IDOK) {
				char s[200];
				GetDlgItemText(hDlg, IDFINDWHAT, s, sizeof(s));
				props->Set("find.what", s);
				GetDlgItemText(hDlg, IDFILES, s, sizeof(s));
				props->Set("find.files", s);
				//GetDlgItemText(hDlg, IDDIRECTORY, s, sizeof(s));
				//props->Set("find.directory", s);
				EndDialog(hDlg,IDOK);
				return TRUE;
			}
		}

		return FALSE;
	}
#endif

	void TideWindow::FindInFiles() {
		SelectionIntoFind();
		props.Set("find.what", findWhat);
		props.Set("find.directory", ".");
#ifdef GTK
		findInFilesDialog = gtk_dialog_new();
		gtk_window_set_policy(GTK_WINDOW(findInFilesDialog), TRUE, TRUE, TRUE);
		gtk_window_set_title(GTK_WINDOW(findInFilesDialog), "Find In Files");

		gtk_signal_connect(GTK_OBJECT(findInFilesDialog),
                   		"destroy", GtkSignalFunc(destroyDialog), 0);

		GtkWidget *table = gtk_table_new(2, 2, FALSE);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog)->vbox),
                   		table, TRUE, TRUE, 0);

		GtkAttachOptions opts = static_cast<GtkAttachOptions>(
                            		GTK_SHRINK | GTK_FILL);

		GtkAttachOptions optse = static_cast<GtkAttachOptions>(
                             		GTK_SHRINK | GTK_FILL | GTK_EXPAND);

		int row = 0;

		GtkWidget *labelFind = gtk_label_new("Find what:");
		gtk_table_attach(GTK_TABLE(table), labelFind, 0, 1,
                 		row, row+1, opts, opts, 5, 5);
		gtk_widget_show(labelFind);

		entryFind = gtk_entry_new();
		gtk_table_attach(GTK_TABLE(table), entryFind, 1, 2,
                 		row, row+1, optse, opts, 5, 5);
		gtk_widget_show(entryFind);
		gtk_entry_set_text(GTK_ENTRY(entryFind), findWhat);
		gtk_entry_select_region(GTK_ENTRY(entryFind), 0, strlen(findWhat));
		gtk_signal_connect(GTK_OBJECT(entryFind),
                   		"activate", GtkSignalFunc(FindInFilesSignal), this);

		row++;

		GtkWidget *labelFiles = gtk_label_new("Files:");
		gtk_table_attach(GTK_TABLE(table), labelFiles, 0, 1,
                 		row, row+1, opts, opts, 5, 5);
		gtk_widget_show(labelFiles);

		entryFiles = gtk_entry_new();
		gtk_table_attach(GTK_TABLE(table), entryFiles, 1, 2,
                 		row, row+1, optse, opts, 5, 5);
		gtk_widget_show(entryFiles);
		gtk_entry_set_text(GTK_ENTRY(entryFiles), props.Get("find.files"));
		gtk_signal_connect(GTK_OBJECT(entryFiles),
                   		"activate", GtkSignalFunc(FindInFilesSignal), this);

		gtk_widget_show(table);

		GtkWidget *buttonFind = gtk_button_new_with_label("Find");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog)->action_area),
                   		buttonFind, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonFind),
                   		"clicked", GtkSignalFunc(FindInFilesSignal), this);
		gtk_widget_show(buttonFind);

		GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog)->action_area),
                   		buttonCancel, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonCancel),
                   		"clicked", GtkSignalFunc(FindInFilesCancelSignal), this);
		gtk_widget_show(buttonCancel);

		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonFind), GTK_CAN_DEFAULT);
		gtk_widget_grab_default(GTK_WIDGET(buttonFind));
		gtk_widget_grab_focus(GTK_WIDGET(entryFind));

		gtk_widget_show(findInFilesDialog);

		gtk_grab_add(findInFilesDialog);
#else
		if (DoDialog(hInstance, "Grep", hwndTide, (FARPROC)GrepDlg, (DWORD)&props) == IDOK) {
			dprintf("asked to find %s %s %s\n", props.Get("find.what"), props.Get("find.files"), props.Get("find.directory"));
			AddCommand(strdup(props.GetNewExpand("find.command", "")));
			if (commandCurrent > 0)
				Execute();
		}
#endif
	}

	void TideWindow::Replace() {
#ifdef GTK
		SelectionIntoFind();
		FindReplace(true);
#else
		if (wFindReplace)
			return;
		SelectionIntoFind();

		memset(&fr, 0, sizeof(fr));
		fr.lStructSize = sizeof(fr);
		fr.hwndOwner = hwndTide;
		fr.hInstance = hInstance;
		fr.Flags = FR_REPLACE;
		fr.lpstrFindWhat = findWhat;
		fr.lpstrReplaceWith = replaceWhat;
		fr.wFindWhatLen = sizeof(findWhat);
		fr.wReplaceWithLen = sizeof(replaceWhat);
		wFindReplace = ReplaceText(&fr);
		replacing = true;
		havefound = false;
		currentDlg = wFindReplace;
#endif
	}

	void TideWindow::ReplaceOnce() {
		if (havefound) {
			SendEditor(EM_REPLACESEL, 0,
           			reinterpret_cast<LPARAM>(static_cast<char *>(replaceWhat)));
			havefound = false;
			//dprintf("Replace <%s> -> <%s>\n", findWhat, replaceWhat);
		}
		FindNext();
	}

	void TideWindow::ReplaceAll() {
		FINDTEXTEX ft;
		ft.chrg.cpMin = 0;
		ft.chrg.cpMax = LengthDocument();
		ft.lpstrText = findWhat;
		ft.chrgText.cpMin = 0;
		ft.chrgText.cpMax = 0;
		int flags = (wholeWord ? FR_WHOLEWORD : 0) | (matchCase ? FR_MATCHCASE : 0);
		int posFind = SendEditor(EM_FINDTEXTEX, flags,
                         		reinterpret_cast<LPARAM>(&ft));
		if (posFind != -1) {
			while (posFind != -1) {
				SendEditor(EM_SETSEL, ft.chrgText.cpMax + 1, ft.chrgText.cpMin);
				SendEditor(EM_REPLACESEL, 0,
           				reinterpret_cast<LPARAM>(static_cast<char *>(replaceWhat)));
				ft.chrg.cpMin = posFind + strlen(replaceWhat) + 1;
				ft.chrg.cpMax = LengthDocument();
				posFind = SendEditor(EM_FINDTEXTEX, flags,
                     				reinterpret_cast<LPARAM>(&ft));
			}
		} else {
			char msg[200];
			strcpy(msg, "No replacements because string \"");
			strcat(msg, findWhat);
			strcat(msg, "\" was not present.");
#ifdef GTK
			MessageBox(hwndTide,msg,"Tide",MB_OK);
#else
			MessageBox(wFindReplace,msg,"Tide",MB_OK);
#endif
		}
		//dprintf("ReplaceAll <%s> -> <%s>\n", findWhat, replaceWhat);
	}

	void TideWindow::OutputAppendString(char *s, int len) {
		if (len == -1)
			len = strlen(s);
		SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
		SendOutput(SCI_ADDTEXT, len, reinterpret_cast<LPARAM>(s));
		SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
	}

#ifdef GTK
	void TideWindow::ContinueExecute() {
		//SendMessage(hwndOutput, EM_SETSEL, 0, 0);
		char buf[256];
		int count = 0;
		count = read(fdFIFO, buf, sizeof(buf)-1 );
		if (count < 0) {
			OutputAppendString(">End Bad\n");
			return;
		}
		if (count == 0) {
			int status;
			int pidWaited = wait(&status);
			char exitmessage[80];
			sprintf(exitmessage, ">Exit code: %d\n", status);
			OutputAppendString(exitmessage);
			// Move selection back to beginning of this run so that F4 will go
			// to first error of this run.
			SendOutput(SCI_GOTOPOS, originalEnd);
			gdk_input_remove(inputHandle);
			close(fdFIFO);
			unlink(resultsFile);
			icmd++;
			if (icmd<commandCurrent && icmd<commandMax) {
				Execute();
			} else {
				icmd=0;
				executing = false;
				for (int ic=0; ic<commandMax; ic++) {
					free(commandToExecute[ic]);
					commandToExecute[ic] = 0;
				}
				commandCurrent = 0;
			}
			return;
		}
		while (count > 0) {
			buf[count] = '\0';
			OutputAppendString(buf);
			if (count < 0) {
				OutputAppendString(">Continue no data\n");
				return;
			}
			count = read(fdFIFO, buf, sizeof(buf)-1 );
		}
	}

	void TideWindow::IOSignal(TideWindow *tidew) {
		tidew->ContinueExecute();
	}
#else
	// ProcessExecute runs a command with redirected input and output streams
	// so the output can be put in a window.
	// It is based upon several usenet posts and a knowledge base article.
	void TideWindow::ProcessExecute() {

		DWORD exitcode = 0;

		SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
		int originalEnd = SendOutput(SCI_GETCURRENTPOS);

		for (int icmd=0;icmd<commandCurrent && icmd < commandMax && exitcode == 0; icmd++) {

			OSVERSIONINFO osv = {sizeof(OSVERSIONINFO)};
			GetVersionEx(&osv);
			bool windows95 = osv.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS;

			SECURITY_ATTRIBUTES sa = {sizeof(SECURITY_ATTRIBUTES)};
			PROCESS_INFORMATION pi = {0};
			HANDLE hPipeWrite  = NULL;
			HANDLE hPipeRead   = NULL;
			HANDLE hWrite2  = NULL;
			HANDLE hRead2   = NULL;
			char buffer[256];
			//dprintf("Execute <%s>\n", command);
			OutputAppendString(">");
			OutputAppendString(commandToExecute[icmd]);
			OutputAppendString("\n");

			sa.bInheritHandle = TRUE;
			sa.lpSecurityDescriptor = NULL;

			SECURITY_DESCRIPTOR sd;
			// If NT make a real security thing to allow inheriting handles
			if (!windows95) {
				InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
				SetSecurityDescriptorDacl(&sd, TRUE, NULL, FALSE);
				sa.nLength = sizeof(SECURITY_ATTRIBUTES);
				sa.lpSecurityDescriptor = &sd;
			}

			// Create pipe for output redirection
			CreatePipe(&hPipeRead,  // read handle
           			&hPipeWrite, // write handle
           			&sa,      // security attributes
           			0      // number of bytes reserved for pipe - 0 default
          			);

			//dprintf("2Execute <%s>\n");
			// Create pipe for input redirection. In this code, you do not
			// redirect the output of the child process, but you need a handle
			// to set the hStdInput field in the STARTUP_INFO struct. For safety,
			// you should not set the handles to an invalid handle.

			CreatePipe(&hRead2,   // read handle
           			&hWrite2,    // write handle
           			&sa,      // security attributes
           			0      // number of bytes reserved for pipe - 0 default
          			);

			//dprintf("3Execute <%s>\n");
			// Make child process use hPipeWrite as standard out, and make
			// sure it does not show on screen.
			STARTUPINFO si = {sizeof(STARTUPINFO)};
			si.dwFlags     = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
			si.wShowWindow = SW_HIDE;
			si.hStdInput   = hWrite2;
			si.hStdOutput  = hPipeWrite;
			si.hStdError   = hPipeWrite;

			bool worked = CreateProcess(
                  			NULL,
                  			commandToExecute[icmd],
                  			NULL, NULL,
                  			TRUE, 0,
                  			NULL, NULL,
                  			&si, &pi);

			if (!worked) {
				OutputAppendString(">Failed to CreateProcess\n");
			}

			/* now that this has been inherited, close it to be safe.
			  You don't want to write to it accidentally */
			CloseHandle(hPipeWrite);

			bool completed = !worked;
			DWORD timeDetectedDeath = 0;
			while (!completed) {

				Sleep(100L);

				if (cancelFlag) {
					TerminateProcess(pi.hProcess, 1);
					completed = true;
					break;
				}

				bool NTOrData = true;

				DWORD bytesRead = 0;

				if (windows95) {
					DWORD bytesAvail = 0;
					if (PeekNamedPipe(hPipeRead, buffer, sizeof(buffer), &bytesRead, &bytesAvail, NULL)) {
						if (0 == bytesAvail) {
							NTOrData = false;
							DWORD dwExitCode = STILL_ACTIVE;
							if (GetExitCodeProcess(pi.hProcess, &dwExitCode)) {
								if (STILL_ACTIVE != dwExitCode) {
									// Process is dead, but wait a second in case there is some output in transit
									if (timeDetectedDeath == 0) {
										timeDetectedDeath = timeGetTime();
									} else {
										if ((timeGetTime() - timeDetectedDeath) > 1000) {
											completed = true; // It's a dead process
										}
									}
								}
							}
						}
					}
				}

				if (!completed && NTOrData) {
					int bTest = ReadFile(
                					hPipeRead,
                					buffer,
                					sizeof(buffer),         // number of bytes to read
                					&bytesRead,
                					NULL         // non-overlapped.
            					);

					if (bTest && bytesRead) {
						// Display the data
						OutputAppendString(buffer, bytesRead);
						UpdateWindow(hwndTide);
					} else {
						completed = true;
					}
				}
			}

			if (worked) {
				WaitForSingleObject(pi.hProcess, INFINITE);
				GetExitCodeProcess(pi.hProcess, &exitcode);
				char exitmessage[80];
				if (exitcode == 0)
					isBuilt = true;
				sprintf(exitmessage, ">Exit code: %d\n", exitcode);
				OutputAppendString(exitmessage);
				CloseHandle(pi.hProcess);
			}
			CloseHandle(hPipeRead);
		}

		// Move selection back to beginning of this run so that F4 will go
		// to first error of this run.
		SendOutput(SCI_GOTOPOS, originalEnd);
		SendMessage(hwndTide, WM_COMMAND, IDM_FINISHEDEXECUTE, 0);
	}

	void ExecThread(void *ptw) {
		TideWindow *tw = reinterpret_cast<TideWindow *>(ptw);
		tw->ProcessExecute();
	}
#endif

	void TideWindow::Execute() {
		SendOutput(SCI_MARKERDELETEALL, 0, 0);
		SendEditor(SCI_MARKERDELETEALL, 0, 0);
#ifndef GTK
		// Ensure the output pane is visible
		if (heightOutput < 20) {
			if (splitVertical)
				heightOutput = NormaliseSplit(300);
			else
				heightOutput = NormaliseSplit(100);
			SizeSubWindows();
			Redraw();
			UpdateWindow(hwndTide);
		}
#endif

		InterlockedExchange(&cancelFlag, 0L);
		executing = true;
		chdir(dirName);
		strcpy(dirNameAtExecute, dirName);

#ifdef GTK
		SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
		originalEnd = SendOutput(SCI_GETCURRENTPOS);

		OutputAppendString(">");
		OutputAppendString(commandToExecute[icmd]);
		OutputAppendString("\n");

		unlink(resultsFile);

		int fdp = mkfifo(resultsFile, S_IRUSR|S_IWUSR);
		if (fdp < 0) {
			OutputAppendString(">Failed to create FIFO\n");
			return;
		}
		close(fdp);

		char commandPlus[1000];
		sprintf(commandPlus, "/bin/sh -c \"%s >%s 2>&1\" &",
        		commandToExecute[icmd], resultsFile);
		pidShell = system(commandPlus);
		int fdFIFO = open(resultsFile, O_RDONLY);
		if (fdFIFO < 0) {
			OutputAppendString(">Failed to open\n");
			return;
		}

		inputHandle = gdk_input_add(fdFIFO, GDK_INPUT_READ,
                            		(GdkInputFunction) IOSignal, this);
#else
		_beginthread(ExecThread, 1024 * 1024, reinterpret_cast<void *>(this));
#endif
	}

#ifndef GTK
	HWND TideWindow::GetWnd() {
		return hwndTide;
	}
#endif

	void TideWindow::Redraw() {
#ifdef GTK
		if (hwndEditor) {
			gtk_widget_queue_draw(hwndEditor);
		}
#else
		InvalidateRect(hwndTide,(LPRECT)NULL,FALSE);
		UpdateWindow(hwndTide);
#endif
	}

	int DecodeMessage(char *cdoc, char *sourcePath,int format) {
		sourcePath[0] = '\0';
		if (format == 1) {
			// Python
			char *startPath = strchr(cdoc, '\"') + 1;
			char *endPath = strchr(startPath, '\"');
			strncpy(sourcePath, startPath, endPath - startPath);
			sourcePath[endPath - startPath] = 0;
			endPath++;
			while (*endPath && !isdigit(*endPath))
				endPath++;
			int sourceNumber = atoi(endPath) - 1;
			return sourceNumber;
		} else if (format == 2) {
			// GCC - look for number followed by colon to be line number
			// This will be preceded by file name
			for (int i=0; cdoc[i]; i++) {
				if (isdigit(cdoc[i]) && cdoc[i+1] == ':') {
					int j = i;
					while (j > 0 && isdigit(cdoc[j-1]))
						j--;
					int sourceNumber = atoi(cdoc + j) - 1;
					strncpy(sourcePath, cdoc, j - 1);
					sourcePath[j-1] = 0;
					return sourceNumber;
				}
			}
		} else if (format == 3) {
			// Visual *
			char *endPath = strchr(cdoc, '(');
			strncpy(sourcePath, cdoc, endPath - cdoc);
			sourcePath[endPath - cdoc] = 0;
			endPath++;
			int sourceNumber = atoi(endPath) - 1;
			return sourceNumber;
		}
		return -1;
	}


	void TideWindow::GoMessage(int dir) {
		int selStart = 0;
		SendOutput(EM_EXGETSEL, reinterpret_cast<WPARAM>(&selStart), 0);
		int curLine = SendOutput(EM_LINEFROMCHAR, selStart);
		int maxLine = SendOutput(EM_GETLINECOUNT, selStart, 0);
		int lookLine = curLine + dir;
		if (lookLine < 0)
			lookLine = maxLine - 1;
		else if (lookLine >= maxLine)
			lookLine = 0;
		while (lookLine != curLine) {
			int startPosLine = SendOutput(EM_LINEINDEX, lookLine, 0);
			int lineLength = SendOutput(EM_LINELENGTH, lookLine, 0);
			dprintf("GOMessage %d %d %d of %d linestart = %d\n", selStart, curLine, lookLine, maxLine, startPosLine);
			char style = SendOutput(SCI_GETSTYLEAT, startPosLine, 0);
			if (style != 0 && style != 4) {
				dprintf("Marker to %d\n", lookLine);
				SendOutput(SCI_MARKERDELETEALL, 0, 0);
				SendOutput(SCI_MARKERDEFINE, 0, SC_MARK_SMALLRECT);
				SendOutput(SCI_MARKERSETFORE, 0, LRGB(0x7f, 0, 0));
				SendOutput(SCI_MARKERSETBACK, 0, LRGB(0xff, 0xff, 0));
				SendOutput(SCI_MARKERADD, lookLine, 0);
				SendOutput(EM_SETSEL, startPosLine, startPosLine);
				char *cdoc = new char[lineLength + 1];
				if (!cdoc)
					return;
				GetRange(hwndOutput, startPosLine, startPosLine + lineLength-1, cdoc);
				char sourcePath[MAX_PATH];
				int sourceLine = DecodeMessage(cdoc, sourcePath, style);
				dprintf("<%s> %d %d\n",sourcePath, sourceLine, lookLine);
				if (sourceLine >= 0) {
					if (0 != strcmp(sourcePath, fileName)) {
						char messagePath[MAX_PATH];
						if (Exists(dirNameAtExecute, sourcePath, messagePath)) {
							Open(messagePath);
						} else if (Exists(dirName, sourcePath, messagePath)) {
							Open(messagePath);
						}
					}
					SendMessage(hwndEditor, SCI_MARKERDELETEALL, 0, 0);
					SendMessage(hwndEditor, SCI_MARKERDEFINE, 0, SC_MARK_CIRCLE);
					SendMessage(hwndEditor, SCI_MARKERSETFORE, 0, LRGB(0x7f, 0, 0));
					SendMessage(hwndEditor, SCI_MARKERSETBACK, 0, LRGB(0xff, 0xff, 0));
					SendMessage(hwndEditor, SCI_MARKERADD, sourceLine, 0);
					int startSourceLine = SendMessage(hwndEditor, EM_LINEINDEX, sourceLine, 0);
					SendMessage(hwndEditor, EM_SETSEL, startSourceLine, startSourceLine);
				}
				delete []cdoc;
				return;
			}
			lookLine += dir;
			if (lookLine < 0)
				lookLine = maxLine - 1;
			else if (lookLine >= maxLine)
				lookLine = 0;
		}
	}

	void TideWindow::StartCallTip() {
		char linebuf[1000];
		int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf), (LPARAM)linebuf);
		int pos = SendEditor(SCI_GETCURRENTPOS);

		int startword=current - 1;
		while (startword > 0 && isalpha(linebuf[startword-1]))
			startword--;
		linebuf[current-1] = '\0';
		char word[203];
		strcpy(word, "fn.");
		strcat(word, linebuf + startword);

		free(functionDefinition);
		functionDefinition = props.GetNewExpand(word, fileName);
		if (functionDefinition && *functionDefinition) {
			SendEditor(SCI_CALLTIPSHOW, pos - strlen(linebuf + startword) - 1, reinterpret_cast<LPARAM>(functionDefinition));
			ContinueCallTip();
		}
	}

	void TideWindow::ContinueCallTip() {
		char linebuf[1000];
		int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf), (LPARAM)linebuf);

		int commas = 0;
		for (int i=0;i<current;i++) {
			if (linebuf[i] == ',')
				commas++;
		}

		int startHighlight = 0;
		while (functionDefinition[startHighlight] && functionDefinition[startHighlight] != '(')
			startHighlight++;
		if (functionDefinition[startHighlight] == '(')
			startHighlight++;
		while (functionDefinition[startHighlight] && commas > 0) {
			if (functionDefinition[startHighlight] == ',' || functionDefinition[startHighlight] == ')')
				commas--;
			startHighlight++;
		}
		if (functionDefinition[startHighlight] == ',' || functionDefinition[startHighlight] == ')')
			startHighlight++;
		int endHighlight = startHighlight;
		if (functionDefinition[endHighlight])
			endHighlight++;
		while (functionDefinition[endHighlight] && functionDefinition[endHighlight] != ',' && functionDefinition[endHighlight] != ')')
			endHighlight++;

		SendEditor(SCI_CALLTIPSETHLT, startHighlight, endHighlight);
	}

	void TideWindow::StartAutoComplete() {
		char linebuf[1000];
		int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf), (LPARAM)linebuf);

		int startword=current - 1;
		while (startword > 0 && isalpha(linebuf[startword-1]))
			startword--;
		linebuf[current-1] = '\0';
		char word[220];
		strcpy(word, "autocompletion.");
		strcat(word, linebuf + startword);

		char *possibles = props.GetNewExpand(word, fileName);
		if (possibles && *possibles)
			SendEditor(SCI_AUTOCSHOW, 0, reinterpret_cast<LPARAM>(possibles));
		free(possibles);
	}

	void TideWindow::SourceChanged() {
		isDirty = true;
		isBuilt = false;
	}

	int TideWindow::GetCurrentLineNumber() {
		int selStart = 0;
		SendEditor(EM_EXGETSEL, (WPARAM)&selStart, 0);
		return SendEditor(EM_LINEFROMCHAR, selStart);
	}

	// Upon a character being added, tide may decide to perform some action
	// such as displaying a completion list.
	void TideWindow::CharAdded(char ch) {
		int selStart = 0;
		int selEnd = 0;
		SendEditor(EM_EXGETSEL, (WPARAM)&selStart, (LPARAM)&selEnd);
		if ((selEnd == selStart) && (selStart > 0)) {
			char style = SendEditor(SCI_GETSTYLEAT, selStart - 1, 0);
			if (style == 0) {
				if (SendEditor(SCI_CALLTIPACTIVE)) {
					if (ch == ')') {
						SendEditor(SCI_CALLTIPCANCEL);
					} else {
						ContinueCallTip();
					}
				} else if (SendEditor(SCI_AUTOCACTIVE)) {
					if (ch == '(') {
						StartCallTip();
					} else if (!isalpha(ch)) {
						SendEditor(SCI_AUTOCCANCEL);
					}
				} else {
					if (ch == '.') {
						StartAutoComplete();
					} else if (ch == '(') {
						StartCallTip();
					} else if (ch == '\r' || ch == '\n') {
						char linebuf[1000];
						int curLine = GetCurrentLineNumber();
						int lineLength = SendEditor(EM_LINELENGTH, curLine);
						//dprintf("[CR] %d len = %d\n", curLine, lineLength);
						if (curLine > 0 && lineLength <= 2) {
							SendEditor(EM_GETLINE, curLine - 1, reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));
							for (int pos = 0;linebuf[pos];pos++) {
								if (linebuf[pos] != ' ' && linebuf[pos] != '\t')
									linebuf[pos] = '\0';
							}
							SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));
						}
					}
				}
			}
		}
	}

#ifdef GTK

	void TideWindow::GotoCancelSignal(GtkWidget *w, TideWindow *tidew) {
		gtk_widget_destroy(GTK_WIDGET(tidew->gotoDialog));
		tidew->gotoDialog = 0;
	}

	void TideWindow::GotoSignal(GtkWidget *w, TideWindow *tidew) {
		char *lineEntry = gtk_entry_get_text(GTK_ENTRY(tidew->gotoEntry));
		int lineNo = atoi(lineEntry);

		tidew->SendEditor(SCI_GOTOLINE, lineNo);

		gtk_widget_destroy(GTK_WIDGET(tidew->gotoDialog));
		tidew->gotoDialog = 0;
	}
#else
	int PASCAL GoLineDlg(HWND hDlg,WORD message,WPARAM wParam,LONG lParam) {
		static int *pLineNo;

		switch (message) {

		case WM_INITDIALOG:
			pLineNo = (int *)lParam;
			return TRUE;

		case WM_CLOSE:
			SendMessage(hDlg,WM_COMMAND,IDCANCEL,NULL);
			break;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg,IDCANCEL);
				return FALSE;
			} else if (LOWORD(wParam) == IDOK) {
				char lineNo[200];
				GetDlgItemText(hDlg, IDGOLINE, lineNo, sizeof(lineNo));
				sscanf(lineNo, "%d", pLineNo);
				EndDialog(hDlg,IDOK);
				return TRUE;
			}
		}

		return FALSE;
	}
#endif

	void TideWindow::GoLineDialog() {
#ifdef GTK
		gotoDialog = gtk_dialog_new();
		gtk_window_set_title(GTK_WINDOW(gotoDialog), "Go To");
		gtk_container_border_width(GTK_CONTAINER(gotoDialog), 0);

		gtk_signal_connect(GTK_OBJECT(gotoDialog),
                   		"destroy", GtkSignalFunc(destroyDialog), 0);

		GtkWidget *table = gtk_table_new(2, 1, FALSE);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog)->vbox),
                   		table, TRUE, TRUE, 0);

		GtkAttachOptions opts = static_cast<GtkAttachOptions>(
                            		GTK_EXPAND | GTK_SHRINK | GTK_FILL);
		GtkWidget *label = gtk_label_new("Go to line:");
		gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, opts, opts, 5, 5);
		gtk_widget_show(label);

		gotoEntry = gtk_entry_new();
		gtk_table_attach(GTK_TABLE(table), gotoEntry, 1, 2, 0, 1, opts, opts, 5, 5);
		gtk_signal_connect(GTK_OBJECT(gotoEntry),
                   		"activate", GtkSignalFunc(GotoSignal), this);
		gtk_widget_grab_focus(GTK_WIDGET(gotoEntry));
		gtk_widget_show(gotoEntry);

		gtk_widget_show(table);

		GtkWidget *buttonGoTo = gtk_button_new_with_label("Go To");
		gtk_signal_connect(GTK_OBJECT(buttonGoTo),
                   		"clicked", GtkSignalFunc(GotoSignal), this);
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonGoTo), GTK_CAN_DEFAULT);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog)->action_area),
                   		buttonGoTo, TRUE, TRUE, 0);
		gtk_widget_grab_default(GTK_WIDGET(buttonGoTo));
		gtk_widget_show(buttonGoTo);

		GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
		gtk_signal_connect(GTK_OBJECT(buttonCancel),
                   		"clicked", GtkSignalFunc(GotoCancelSignal), this);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog)->action_area),
                   		buttonCancel, TRUE, TRUE, 0);
		gtk_widget_show(buttonCancel);

		gtk_widget_show(gotoDialog);

		gtk_grab_add(gotoDialog);
#else
		int lineNo = 0;
		if (DoDialog(hInstance, "GoLine", hwndTide, (FARPROC)GoLineDlg, (DWORD)&lineNo) == IDOK) {
			//dprintf("asked to go to %d\n", lineNo);
			SendEditor(SCI_GOTOLINE, lineNo);
		}
		SetFocus(hwndEditor);
#endif
	}

#ifdef GTK

	void TideWindow::FindReplaceGrabFields() {
		char *findEntry = gtk_entry_get_text(GTK_ENTRY(entryFind));
		strncpy(findWhat, findEntry, sizeof(findWhat));
		if (entryReplace) {
			char *replaceEntry = gtk_entry_get_text(GTK_ENTRY(entryReplace));
			strncpy(replaceWhat, replaceEntry, sizeof(replaceWhat));
		}
		wholeWord = GTK_TOGGLE_BUTTON(toggleWord)->active;
		matchCase = GTK_TOGGLE_BUTTON(toggleCase)->active;
	}

	void TideWindow::FRCancelSignal(GtkWidget *w, TideWindow *tidew) {
		gtk_widget_destroy(GTK_WIDGET(tidew->wFindReplace));
		tidew->wFindReplace = 0;
	}

	void TideWindow::FRFindSignal(GtkWidget *w, TideWindow *tidew) {
		tidew->FindReplaceGrabFields();
		if (tidew->findWhat[0]) {
			tidew->FindNext();
		}
	}

	void TideWindow::FRReplaceSignal(GtkWidget *w, TideWindow *tidew) {
		tidew->FindReplaceGrabFields();
		tidew->ReplaceOnce();
	}

	void TideWindow::FRReplaceAllSignal(GtkWidget *w, TideWindow *tidew) {
		tidew->FindReplaceGrabFields();
		tidew->ReplaceAll();
		gtk_widget_destroy(GTK_WIDGET(tidew->wFindReplace));
		tidew->wFindReplace = 0;
	}
#endif

	void TideWindow::FindReplace(bool replace) {
		replacing = replace;
#ifdef GTK
		wFindReplace = gtk_dialog_new();
		gtk_window_set_policy(GTK_WINDOW(wFindReplace), TRUE, TRUE, TRUE);
		gtk_window_set_title(GTK_WINDOW(wFindReplace),
                     		replace ? "Replace" : "Find");

		gtk_signal_connect(GTK_OBJECT(wFindReplace),
                   		"destroy", GtkSignalFunc(destroyDialog), 0);

		GtkWidget *table = gtk_table_new(2, replace ? 4 : 3, FALSE);
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace)->vbox),
                   		table, TRUE, TRUE, 0);

		GtkAttachOptions opts = static_cast<GtkAttachOptions>(
                            		GTK_SHRINK | GTK_FILL);

		GtkAttachOptions optse = static_cast<GtkAttachOptions>(
                             		GTK_SHRINK | GTK_FILL | GTK_EXPAND);

		int row = 0;

		GtkWidget *labelFind = gtk_label_new("Find:");
		gtk_table_attach(GTK_TABLE(table), labelFind, 0, 1,
                 		row, row+1, opts, opts, 5, 5);
		gtk_widget_show(labelFind);

		entryFind = gtk_entry_new();
		gtk_table_attach(GTK_TABLE(table), entryFind, 1, 2,
                 		row, row+1, optse, opts, 5, 5);
		gtk_widget_show(entryFind);
		gtk_entry_set_text(GTK_ENTRY(entryFind), findWhat);
		gtk_entry_select_region(GTK_ENTRY(entryFind), 0, strlen(findWhat));
		gtk_signal_connect(GTK_OBJECT(entryFind),
                   		"activate", GtkSignalFunc(FRFindSignal), this);

		row++;

		if (replace) {
			GtkWidget *labelReplace = gtk_label_new("Replace:");
			gtk_table_attach(GTK_TABLE(table), labelReplace, 0, 1,
                 			row, row+1, opts, opts, 5, 5);
			gtk_widget_show(labelReplace);

			entryReplace = gtk_entry_new();
			gtk_table_attach(GTK_TABLE(table), entryReplace, 1, 2,
                 			row, row+1, optse, opts, 5, 5);
			gtk_widget_show(entryReplace);
			gtk_signal_connect(GTK_OBJECT(entryReplace),
                   			"activate", GtkSignalFunc(FRFindSignal), this);

			row++;
		} else {
			entryReplace = 0;
		}

		// Case Sensitive
		toggleCase = gtk_check_button_new_with_label("Case Sensitive");
		gtk_table_attach(GTK_TABLE(table), toggleCase, 0, 2, row, row+1, opts, opts, 3, 0);
		gtk_widget_show(toggleCase);
		row++;

		// Whole Word
		toggleWord = gtk_check_button_new_with_label("Whole Word");
		gtk_table_attach(GTK_TABLE(table), toggleWord, 0, 2, row, row+1, opts, opts, 3, 0);
		gtk_widget_show(toggleWord);

		gtk_widget_show(table);

		GtkWidget *buttonFind = gtk_button_new_with_label("Find");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace)->action_area),
                   		buttonFind, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonFind),
                   		"clicked", GtkSignalFunc(FRFindSignal), this);
		gtk_widget_show(buttonFind);

		if (replace) {
			GtkWidget *buttonReplace = gtk_button_new_with_label("Replace");
			gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace)->action_area),
                   			buttonReplace, TRUE, TRUE, 0);
			gtk_signal_connect(GTK_OBJECT(buttonReplace),
                   			"clicked", GtkSignalFunc(FRReplaceSignal), this);
			gtk_widget_show(buttonReplace);

			GtkWidget *buttonReplaceAll = gtk_button_new_with_label("Replace All");
			gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace)->action_area),
                   			buttonReplaceAll, TRUE, TRUE, 0);
			gtk_signal_connect(GTK_OBJECT(buttonReplaceAll),
                   			"clicked", GtkSignalFunc(FRReplaceAllSignal), this);
			gtk_widget_show(buttonReplaceAll);
		}

		GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace)->action_area),
                   		buttonCancel, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonCancel),
                   		"clicked", GtkSignalFunc(FRCancelSignal), this);
		gtk_widget_show(buttonCancel);

		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonFind), GTK_CAN_DEFAULT);
		gtk_widget_grab_default(GTK_WIDGET(buttonFind));
		gtk_widget_grab_focus(GTK_WIDGET(entryFind));

		gtk_widget_show(wFindReplace);

		gtk_grab_add(wFindReplace);
#else
#endif
	}

	void TideWindow::AddCommand(char *cmd) {
		if (cmd) {
			dprintf("AddCOmmand %s\n",cmd);
			if (cmd[0]) {
				commandToExecute[commandCurrent] = cmd;
				commandCurrent++;
			} else {
				free(cmd);
			}
		}
	}

	void TideWindow::Command(WPARAM wParam, LPARAM lParam) {
		int lineNo = 0;

		switch (LOWORD(wParam)) {
		case IDM_NEW:
			if (SaveIfUnsure() != IDCANCEL) {
				New();
			}
			break;
		case IDM_OPEN:
			if (SaveIfUnsure() != IDCANCEL) {
				Open();
				SetFocus(hwndEditor);
			}
			break;
		case IDM_CLOSE:
			if (SaveIfUnsure() != IDCANCEL) {
				DropFileStackTop();
				StackMenu(0);
				// If none left, New()
				SetFocus(hwndEditor);
			}
			break;
		case IDM_SAVE:
			Save();
			SetFocus(hwndEditor);
			break;
		case IDM_SAVEAS:
			SaveAs();
			SetFocus(hwndEditor);
			break;
		case IDM_REVERT:
			break;
		case IDM_PRINT:
			MessageBox(hwndTide,"Print not implemented.","Tide",MB_OK);
			SetFocus(hwndEditor);
			break;
		case IDM_PRINTSETUP:
			MessageBox(hwndTide,"Printer Setup not implemented.","Tide",MB_OK);
			SetFocus(hwndEditor);
			break;
		case IDM_ABOUT:
#ifdef GTK
			MessageBox(hwndTide,"Tide 0.90\nby Neil Hodgson neilh@hare.net.au.","Tide",MB_OK);
#else
			DoDialog(hInstance,"About",hwndTide,NULL,NULL);
			SetFocus(hwndEditor);
#endif
			break;
		case IDM_QUIT:
			if (SaveIfUnsure() != IDCANCEL) {
#ifdef GTK
				gtk_exit(0);
#else
				SendMessage(hwndTide,WM_CLOSE,0,0L);
#endif
			}
			break;
		case IDM_NEXTFILE:
			if (SaveIfUnsure() != IDCANCEL) {
				StackMenu(1);
			}
			break;

		case IDM_UNDO:
			SendEditor(WM_UNDO);
			break;
		case IDM_REDO:
			SendEditor(SCI_REDO);
			break;

		case IDM_CUT:
			SendEditor(WM_CUT);
			break;
		case IDM_COPY:
			SendEditor(WM_COPY);
			break;
		case IDM_PASTE:
			SendEditor(WM_PASTE);
			break;
		case IDM_CLEAR:
			SendEditor(WM_CLEAR);
			break;
		case IDM_SELECTALL:
			SendEditor(SCI_SELECTALL);
			break;

		case IDM_FIND:
			Find();
			break;

		case IDM_FINDNEXT:
			FindNext();
			break;

		case IDM_FINDINFILES:
			FindInFiles();
			break;

		case IDM_REPLACE:
			Replace();
			break;

		case IDM_GOTO:
			GoLineDialog();
			break;

		case IDM_SPLITVERTICAL:
			splitVertical = !splitVertical;
			heightOutput = NormaliseSplit(heightOutput);
			SizeSubWindows();
			Redraw();
			break;

		case IDM_UPRCASE:
			//dprintf("IDM_UPRCASE\n");
			break;
		case IDM_LWRCASE:
			//dprintf("IDM_LWRCASE\n");
			break;
		case IDM_VIEWSPACE: {
				int viewWS = SendEditor(SCI_GETVIEWWS, 0, 0);
				SendEditor(SCI_SETVIEWWS, !viewWS);
				Redraw();
			}
			break;

		case IDM_COMPILE: {
				if (SaveIfUnsure() != IDCANCEL) {
					AddCommand(props.GetNewExpand("command.compile.", fileName));
					if (commandCurrent > 0)
						Execute();
				}
			}
			break;

		case IDM_BUILD: {
				if (SaveIfUnsure() != IDCANCEL) {
					AddCommand(props.GetNewExpand("command.build.", fileName));
					if (commandCurrent > 0)
						Execute();
				}
			}
			break;

		case IDM_GO: {
				if (SaveIfUnsure() != IDCANCEL) {
					if (!isBuilt)
						AddCommand(props.GetNewExpand("command.go.needs.", fileName));
					AddCommand(props.GetNewExpand("command.go.", fileName));
					if (commandCurrent > 0)
						Execute();
				}
			}
			break;

#ifndef GTK
		case IDM_STOPEXECUTE:
			InterlockedExchange(&cancelFlag, 1L);
			break;

		case IDM_FINISHEDEXECUTE: {
				executing = false;
				for (int icmd=0; icmd<commandMax; icmd++) {
					free(commandToExecute[icmd]);
					commandToExecute[icmd] = 0;
				}
				commandCurrent = 0;
			}
			break;
#endif

		case IDM_NEXTMSG:
			GoMessage(1);
			break;

		case IDM_PREVMSG:
			GoMessage(-1);
			break;

		case IDM_OPENLOCALPROPERTIES:
			OpenProperties(true);
			SetFocus(hwndEditor);
			break;

		case IDM_OPENGLOBALPROPERTIES:
			OpenProperties(false);
			SetFocus(hwndEditor);
			break;

		case IDM_STEP: {
				if (0 == strcmp(language, "pyro")) {
					int posns[] = {7, 8, 9, 10, 11, 10, 11, 10, 11};
					SendEditor(SCI_MARKERDELETE, posns[currentStep], 2);
					currentStep++;
					if (currentStep == 9)
						currentStep = 0;
					SendEditor(SCI_MARKERADD, posns[currentStep], 2);
					Redraw();
				}
			}
			break;

		case IDM_TABTIMMY: {
				int lengthDoc = GetWindowTextLength(hwndEditor);
				char *cdoc = new char[lengthDoc];
				//dprintf("Tabtimmy size=%d ptr=%x\n", lengthDoc, cdoc);
				// Have to use SendMessage rather than GetWindowText as GetWindowText truncates
				// length parameter to 16 bits
				SendEditor(WM_GETTEXT, lengthDoc, reinterpret_cast<LPARAM>(cdoc));
				TabTimmy(cdoc, lengthDoc, hwndEditor);
				Redraw();
				delete []cdoc;
			}
			break;

		case IDM_SELECTIONMARGIN:
			selMargin = !selMargin;
			SendEditor(SCI_SETMARGINWIDTH, selMargin ? 20 : 0);
			Redraw();
			break;

		case IDM_BUFFEREDDRAW:
			bufferedDraw = !bufferedDraw;
			SendEditor(SCI_SETBUFFEREDDRAW, bufferedDraw);
			break;

		case IDM_SRCWIN: {
				int cmd = HIWORD(wParam);
				if (cmd == EN_CHANGE) {
					SourceChanged();
				}
			}
			break;

		default:
			if ((wParam >= fileStackCmdID) &&
    				(wParam < fileStackCmdID + fileStackMax)) {
				if (SaveIfUnsure() != IDCANCEL) {
					StackMenu(wParam-fileStackCmdID);
				}
			}
			break;

		}
	}

	void TideWindow::Notify(SCNotification *notification) {
		switch (notification->nmhdr.code) {
		case SCN_STYLENEEDED: {
				if (notification->nmhdr.idFrom == IDM_SRCWIN) {
					int endStyled = SendEditor(SCI_GETENDSTYLED);
					int lineEndStyled = SendEditor(EM_LINEFROMCHAR, endStyled);
					endStyled = SendEditor(EM_LINEINDEX, lineEndStyled);
					Colourise(endStyled, notification->position);
				} else {
					int endStyled = SendOutput(SCI_GETENDSTYLED);
					int lineEndStyled = SendOutput(EM_LINEFROMCHAR, endStyled);
					endStyled = SendOutput(EM_LINEINDEX, lineEndStyled);
					Colourise(endStyled, notification->position, false);
				}
			}
			break;

		case SCN_CHARADDED:
			CharAdded(notification->ch);
			break;
#ifdef GTK
		case SCN_KEY: {
				int mods = 0;
				if (notification->modifiers & SHIFT_PRESSED)
					mods |= GDK_SHIFT_MASK;
				if (notification->modifiers & LEFT_CTRL_PRESSED)
					mods |= GDK_CONTROL_MASK;
				//dprintf("SCN_KEY: %d %d\n", notification->ch, mods);
				gtk_accel_group_activate(accelGroup, notification->ch,
                         				static_cast<GdkModifierType>(mods));
			}
			break;
#endif
		}
	}

	void TideWindow::DoPop(HWND hWnd,POINT p) {
	}

#ifdef GTK
	void CheckAMenuItem(GtkWidget *hWnd,const char *path,BOOL val) {
	}

	void EnableAMenuItem(GtkWidget *hWnd,const char *path,BOOL val) {
	}
#else
	void CheckAMenuItem(HWND hWnd,WORD wIDCheckItem,BOOL val) {
		if (val)
			CheckMenuItem(GetMenu(hWnd),wIDCheckItem,MF_CHECKED | MF_BYCOMMAND);
		else
			CheckMenuItem(GetMenu(hWnd),wIDCheckItem,MF_UNCHECKED | MF_BYCOMMAND);
	}

	void EnableAMenuItem(HWND hWnd,WORD wIDCheckItem,BOOL val) {
		if (val)
			EnableMenuItem(GetMenu(hWnd),wIDCheckItem,MF_ENABLED | MF_BYCOMMAND);
		else
			EnableMenuItem(GetMenu(hWnd),wIDCheckItem,MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
	}
#endif

	void TideWindow::CheckMenus() {
#ifdef GTK
		EnableAMenuItem(hwndTide,"/File/Save",isDirty);
		EnableAMenuItem(hwndTide,"/Edit/Undo",SendEditor(EM_CANUNDO));
		EnableAMenuItem(hwndTide,"/Edit/Paste",SendEditor(EM_CANPASTE));
		//CheckAMenuItem(hwndTide,"/"IDM_SPLITVERTICAL,splitVertical);
		CheckAMenuItem(hwndTide,"/Edit/View Whitespace",SendEditor(SCI_GETVIEWWS));
		CheckAMenuItem(hwndTide,"/Pyro/Selection Margin",selMargin);
		CheckAMenuItem(hwndTide,"/Pyro/Buffered Drawing",bufferedDraw);
		EnableAMenuItem(hwndTide,"/Tools/Compile",!executing);
		EnableAMenuItem(hwndTide,"/Tools/Build",!executing);
		EnableAMenuItem(hwndTide,"/Tools/Go",!executing);
		//EnableAMenuItem(hwndTide,IDM_STOPEXECUTE,executing);
#else
		EnableAMenuItem(hwndTide,IDM_SAVE,isDirty);
		EnableAMenuItem(hwndTide,IDM_UNDO,SendEditor(EM_CANUNDO));
		EnableAMenuItem(hwndTide,IDM_PASTE,SendEditor(EM_CANPASTE));
		CheckAMenuItem(hwndTide,IDM_SPLITVERTICAL,splitVertical);
		CheckAMenuItem(hwndTide,IDM_VIEWSPACE,SendEditor(SCI_GETVIEWWS));
		CheckAMenuItem(hwndTide,IDM_SELECTIONMARGIN,selMargin);
		CheckAMenuItem(hwndTide,IDM_BUFFEREDDRAW,bufferedDraw);
		EnableAMenuItem(hwndTide,IDM_COMPILE,!executing);
		EnableAMenuItem(hwndTide,IDM_BUILD,!executing);
		EnableAMenuItem(hwndTide,IDM_GO,!executing);
		EnableAMenuItem(hwndTide,IDM_STOPEXECUTE,executing);
#endif
	}

	void TideWindow::SizeSubWindows() {
		RECT rcClient;
		GetClientRect(hwndTide, &rcClient);
		int w = rcClient.right - rcClient.left + 1;
		int h = rcClient.bottom - rcClient.top + 1;
		heightOutput = NormaliseSplit(heightOutput);
		if (splitVertical) {
			SetWindowPos(hwndEditor, 0, 0, 0, w - heightOutput - heightBar, h-1, 0);
			SetWindowPos(hwndOutput, 0, w - heightOutput, 0, heightOutput-1, h-1, 0);
		} else {
			SetWindowPos(hwndEditor, 0, 0, 0, w-1, h - heightOutput - heightBar, 0);
			SetWindowPos(hwndOutput, 0, 0, h - heightOutput, w-1, heightOutput-1, 0);
		}
	}

	static POINT PointFromLparam(LPARAM lParam) {
		POINT pt;
		pt.x = (int)(short)LOWORD(lParam);
		pt.y = (int)(short)HIWORD(lParam);
		return pt;
	}

	// Ensure that a splitter bar position is inside the main window
	int TideWindow::NormaliseSplit(int splitPos) {
		RECT rcClient = {0, 0, 200, 200};
		GetClientRect(hwndTide, &rcClient);
		int w = rcClient.right - rcClient.left + 1;
		int h = rcClient.bottom - rcClient.top + 1;
		if (splitPos < 20)
			splitPos = 0;
		if (splitVertical) {
			if (splitPos > w - heightBar - 20)
				splitPos = w - heightBar;
		} else {
			if (splitPos > h - heightBar - 20)
				splitPos = h - heightBar;
		}
		return splitPos;
	}

	void TideWindow::MoveSplit(POINT ptNewDrag) {
		RECT rcClient = {0, 0, 200, 200};
		GetClientRect(hwndTide, &rcClient);
		int w = rcClient.right - rcClient.left + 1;
		int h = rcClient.bottom - rcClient.top + 1;
		int newHeightOutput = heightOutputStartDrag + (ptStartDrag.y - ptNewDrag.y);
		if (splitVertical)
			newHeightOutput = heightOutputStartDrag + (ptStartDrag.x - ptNewDrag.x);
		newHeightOutput = NormaliseSplit(newHeightOutput);
		if (heightOutput != newHeightOutput) {
			heightOutput = newHeightOutput;
			SizeSubWindows();
			Redraw();
		}
	}

#ifdef GTK
	void TideWindow::QuitSignal(TideWindow *tidew, guint action, GtkWidget *w) {
		tidew->Command(IDM_QUIT, 0);
	}

	void TideWindow::MenuSignal(TideWindow *tidew, guint action, GtkWidget *w) {
		//dprintf("action %x %d %x \n",tidew, action, w);
		tidew->Command(action, 0);
	}

	void TideWindow::CommandSignal(GtkWidget *w, gint wParam, gpointer lParam, TideWindow *tidew) {
		//dprintf("Command: %x %x %x %x\n", w, wParam, lParam, tidew);
		tidew->Command(wParam, reinterpret_cast<LPARAM>(lParam));
	}

	void TideWindow::NotifySignal(GtkWidget *w, gint wParam, gpointer lParam, TideWindow *tidew) {
		//dprintf("Notify: %x %x %x %x\n", w, wParam, lParam, tidew);
		tidew->Notify(reinterpret_cast<SCNotification *>(lParam));
	}

	void TideWindow::Run(char *cmdLine) {

		GtkItemFactoryCallback menuSig = GtkItemFactoryCallback(MenuSignal);
		GtkItemFactoryEntry menuItems[] = {
    		{"/_File", 		NULL, 		NULL, 0, "<Branch>"
    		},
    		{"/File/_New", 	"<control>N", 	menuSig, IDM_NEW, 0},
    		{"/File/_Open", 	"<control>O", 	menuSig, IDM_OPEN, 0},
    		{"/File/_Close", 	NULL, 		menuSig, IDM_CLOSE, 0},
    		{"/File/_Save", 	"<control>S", 	menuSig, IDM_SAVE, 0},
    		{"/File/Save _As", 	NULL, 		menuSig, IDM_SAVEAS, 0},
    		{"/File/sep1", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/File/_Quit", 		"<control>Q", 	menuSig, IDM_QUIT, 0},

    		{"/_Edit", 		NULL, 		NULL, 0, "<Branch>"},
    		{"/Edit/_Undo", 	"<control>Z", 	menuSig, IDM_UNDO, 0},
    		{"/Edit/_Redo", 	"<control>Y", 	menuSig, IDM_REDO, 0},
    		{"/Edit/sep1", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Edit/Cu_t", 		"<control>X", 	menuSig, IDM_CUT, 0},
    		{"/Edit/_Copy", 	"<control>C", 	menuSig, IDM_COPY, 0},
    		{"/Edit/_Paste", 	"<control>V", 	menuSig, IDM_PASTE, 0},
    		{"/Edit/_Delete", 	"Del", 		menuSig, IDM_CLEAR, 0},
    		{"/Edit/Select _All", 	"", 		menuSig, IDM_SELECTALL, 0},
    		{"/Edit/sep2", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Edit/_Find...", 	"<control>F", 	menuSig, IDM_FIND, 0},
    		{"/Edit/Find _Next", 	"F3", 		menuSig, IDM_FINDNEXT, 0},
    		{"/Edit/F_ind in Files...",	"", 		menuSig, IDM_FINDINFILES, 0},
    		{"/Edit/R_eplace", 	"<control>H", 	menuSig, IDM_REPLACE, 0},
    		{"/Edit/sep3", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Edit/_Go To", 	"<control>G", 	menuSig, IDM_GOTO, 0},
    		//{"/Edit/Vertical Split", 	"", 		menuSig, IDM_SPLITVERTICAL, NULL},
    		{"/Edit/sep4", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Edit/View _Whitespace", "<control>8", 	menuSig, IDM_VIEWSPACE, 0},

    		{"/_Pyro", 		NULL, 		NULL, 0, "<Branch>"},
    		{"/Pyro/TabTimmy", 	"", 		menuSig, IDM_TABTIMMY, 0},
    		{"/Pyro/Selection Margin", "", 		menuSig, IDM_SELECTIONMARGIN, 0},
    		{"/Pyro/Buffered Drawing", "", 		menuSig, IDM_BUFFEREDDRAW, 0},
    		{"/Pyro/Step", 		"F8", 		menuSig, IDM_STEP, 0},

    		{"/_Tools", 		NULL, 		NULL, 0, "<Branch>"},
    		{"/Tools/_Compile", 	"<control>F7", 	menuSig, IDM_COMPILE, 0},
    		{"/Tools/_Build", 	"F7", 		menuSig, IDM_BUILD, 0},
    		{"/Tools/_Go", 		"F5", 		menuSig, IDM_GO, 0},
    		//{"/Tools/_Stop Executing",	"<control>.", 	menuSig, IDM_STOPEXECUTE, NULL},
    		{"/Tools/sep1", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Tools/Next Message", 	"F4", 		menuSig, IDM_NEXTMSG, 0},
    		{"/Tools/Previous Message", "<shift>F4", 	menuSig, IDM_PREVMSG, 0},
    		{"/Tools/sep2", 		NULL, 		NULL, 0, "<Separator>"},
    		{"/Tools/Open _Local Options File", "", 	menuSig, IDM_OPENLOCALPROPERTIES, 0},
    		{"/Tools/O_pen Global Options File", "", 	menuSig, IDM_OPENGLOBALPROPERTIES, 0},

    		{"/_Help", 		NULL, 		NULL, 0, "<Branch>"},
    		{"/Help/About Tide", 	"", 		menuSig, IDM_ABOUT, 0},
		};

		char *gthis = reinterpret_cast<char *>(this);

		window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		hwndTide = window;
		GTK_WIDGET_UNSET_FLAGS(window, GTK_CAN_FOCUS);
		gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, FALSE);

		gtk_widget_set_events (window,
                       		GDK_EXPOSURE_MASK
                       		| GDK_LEAVE_NOTIFY_MASK
                       		| GDK_BUTTON_PRESS_MASK
                       		| GDK_BUTTON_RELEASE_MASK
                      		);
		gtk_signal_connect(GTK_OBJECT(window), "destroy",
                   		GTK_SIGNAL_FUNC(QuitSignal), gthis);
		gtk_window_set_title(GTK_WINDOW(window), "Tide");
		gtk_widget_set_usize(GTK_WIDGET(window), 600, 700);

		GtkWidget *boxMain = gtk_vbox_new(FALSE,1);
		gtk_container_add(GTK_CONTAINER(window), boxMain);
		GTK_WIDGET_UNSET_FLAGS(boxMain, GTK_CAN_FOCUS);

		int nItems = sizeof(menuItems) / sizeof(menuItems[0]);
		accelGroup = gtk_accel_group_new();
		itemFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelGroup);
		gtk_item_factory_create_items(itemFactory, nItems, menuItems, gthis);
		gtk_accel_group_attach(accelGroup, GTK_OBJECT(window));

		gtk_box_pack_start(GTK_BOX(boxMain),
                   		gtk_item_factory_get_widget(itemFactory, "<main>"),
                   		FALSE, FALSE, 0);

		GtkWidget *vpaned = gtk_vpaned_new();
		GTK_WIDGET_UNSET_FLAGS(vpaned, GTK_CAN_FOCUS);
		gtk_paned_gutter_size(GTK_PANED(vpaned),8);
		gtk_box_pack_start(GTK_BOX(boxMain), vpaned, TRUE, TRUE, 0);

		hwndEditor = scintilla_new();
		scintilla_set_id(SCINTILLA(hwndEditor), IDM_SRCWIN);
		gtk_paned_add1(GTK_PANED(vpaned), hwndEditor);
		gtk_widget_set_usize(hwndEditor, 600, 600);
		gtk_signal_connect(GTK_OBJECT(hwndEditor), "command",
                   		GtkSignalFunc(CommandSignal), this);
		gtk_signal_connect(GTK_OBJECT(hwndEditor), "notify",
                   		GtkSignalFunc(NotifySignal), this);

		hwndOutput = scintilla_new();
		scintilla_set_id(SCINTILLA(hwndOutput), IDM_RUNWIN);
		gtk_paned_add2(GTK_PANED(vpaned), hwndOutput);
		gtk_widget_set_usize(hwndOutput, 600, 100);
		gtk_signal_connect(GTK_OBJECT(hwndOutput), "command",
                   		GtkSignalFunc(CommandSignal), this);
		gtk_signal_connect(GTK_OBJECT(hwndOutput), "notify",
                   		GtkSignalFunc(NotifySignal), this);

		SendOutput(SCI_SETMARGINWIDTH, 0, 0);

		Open(cmdLine,true);

		gtk_widget_show_all(window);
		SetFocus(hwndEditor);
		gtk_main();
	}

	int main(int argc, char *argv[]) {
		gtk_init(&argc, &argv);

		TideWindow tide;
		if (argc > 1) {
			//dprintf("args: %d %s\n", argc, argv[1]);
			tide.Run(argv[1]);
		} else {
			tide.Run("");
		}

		return 0;
	}
#else
	long TideWindow::WndProc(WORD iMessage,WPARAM wParam,LPARAM lParam) {
		dprintf("start wnd proc %x %x\n",iMessage, hwndTide);
		switch (iMessage) {

		case WM_CREATE:
			hwndEditor = CreateWindow(
                 			"Scintilla",
                 			"Source",
                 			WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_CLIPCHILDREN,
                 			0, 0,
                 			100, 100,
                 			hwndTide,
                 			(HMENU)IDM_SRCWIN,
                 			hInstance,
                 			0);
			if (!hwndEditor)
				exit(FALSE);
			ShowWindow(hwndEditor, SW_SHOWNORMAL);
			SendEditor(SCI_ASSIGNCMDKEY, VK_RETURN, SCI_NEWLINE);
			SendEditor(SCI_ASSIGNCMDKEY, VK_TAB, SCI_TAB);
			SendEditor(SCI_ASSIGNCMDKEY, VK_TAB | (SHIFT_PRESSED<<16), SCI_BACKTAB);
			SetFocus(hwndEditor);

			hwndOutput = CreateWindow(
                 			"Scintilla",
                 			"Run",
                 			WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_CLIPCHILDREN,
                 			0, 0,
                 			100, 100,
                 			hwndTide,
                 			(HMENU)IDM_RUNWIN,
                 			hInstance,
                 			0);
			if (!hwndOutput)
				exit(FALSE);
			ShowWindow(hwndOutput, SW_SHOWNORMAL);
			//SendOutput(SCI_ADDTEXT, 6, (LPARAM)">>>>\r\n");
			// No selection margin on output window
			SendOutput(SCI_SETMARGINWIDTH, 0, 0);
			SendOutput(SCI_ASSIGNCMDKEY, VK_RETURN, SCI_NEWLINE);
			SendOutput(SCI_ASSIGNCMDKEY, VK_TAB, SCI_TAB);
			SendOutput(SCI_ASSIGNCMDKEY, VK_TAB | (SHIFT_PRESSED<<16), SCI_BACKTAB);
			break;

		case WM_PAINT: {
				PAINTSTRUCT ps;
				BeginPaint(hwndTide,&ps);
				RECT rcClient = {0,0,0,0};
				GetClientRect(hwndTide, &rcClient);
				int heightClient = rcClient.bottom - rcClient.top;
				int widthClient = rcClient.right - rcClient.left;

				HBRUSH brushSurface = CreateSolidBrush(GetSysColor(COLOR_3DFACE));
				FillRect(ps.hdc, &rcClient, brushSurface);
				DeleteObject(brushSurface);

				HPEN penHilight = CreatePen(0,1,GetSysColor(COLOR_3DHILIGHT));
				HPEN penOld = (HPEN)SelectObject(ps.hdc, penHilight);
				HPEN penShadow = CreatePen(0,1,GetSysColor(COLOR_3DSHADOW));
				HPEN penDark = CreatePen(0,1,GetSysColor(COLOR_3DDKSHADOW));

				if (splitVertical) {
					MoveToEx(ps.hdc, widthClient - (heightOutput + heightBar - 1), 0, 0);
					LineTo(ps.hdc, widthClient - (heightOutput + heightBar - 1), heightClient);

					SelectObject(ps.hdc, penShadow);
					MoveToEx(ps.hdc, widthClient - (heightOutput + 2), 0, 0);
					LineTo(ps.hdc, widthClient - (heightOutput + 2), heightClient);

					SelectObject(ps.hdc, penDark);
					MoveToEx(ps.hdc, widthClient - (heightOutput + 1), 0, 0);
					LineTo(ps.hdc, widthClient - (heightOutput + 1), heightClient);
				} else {
					MoveToEx(ps.hdc, 0, heightClient - (heightOutput + heightBar - 1) , 0);
					LineTo(ps.hdc, widthClient, heightClient - (heightOutput + heightBar - 1));

					SelectObject(ps.hdc, penShadow);
					MoveToEx(ps.hdc, 0, heightClient - (heightOutput + 2), 0);
					LineTo(ps.hdc, widthClient, heightClient - (heightOutput + 2));

					SelectObject(ps.hdc, penDark);
					MoveToEx(ps.hdc, 0, heightClient - (heightOutput + 1), 0);
					LineTo(ps.hdc, widthClient, heightClient - (heightOutput + 1));
				}

				SelectObject(ps.hdc, penOld);
				DeleteObject(penDark);
				DeleteObject(penShadow);
				DeleteObject(penHilight);
				EndPaint(hwndTide,&ps);
			}
			break;

		case WM_COMMAND:
			Command(wParam, lParam);
			break;

		case WM_NOTIFY:
			Notify(reinterpret_cast<SCNotification *>(lParam));
			break;

		case WM_KEYDOWN:
			dprintf("keydown %d %x %x\n",iMessage, wParam, lParam);
			break;

		case WM_KEYUP:
			dprintf("keyup %d %x %x\n",iMessage, wParam, lParam);
			break;

		case WM_SIZE:
			SizeSubWindows();
			break;

		case WM_GETMINMAXINFO:
			return DefWindowProc(hwndTide,iMessage,wParam,lParam);
			break;

		case WM_LBUTTONDOWN:
			ptStartDrag = PointFromLparam(lParam);
			capturedMouse = true;
			heightOutputStartDrag = heightOutput;
			SetCapture(hwndTide);
			dprintf("Click %x %x\n", wParam, lParam);
			break;

		case WM_MOUSEMOVE:
			if (capturedMouse) {
				POINT ptNewDrag = PointFromLparam(lParam);
				MoveSplit(ptNewDrag);
			}
			break;

		case WM_LBUTTONUP:
			if (capturedMouse) {
				POINT ptNewDrag = PointFromLparam(lParam);
				MoveSplit(ptNewDrag);
				capturedMouse = false;
				ReleaseCapture();
			}
			break;

		case WM_SETCURSOR:
			if (LOWORD(lParam) == HTCLIENT) {
				POINT ptCursor;
				GetCursorPos(&ptCursor);
				RECT rcScintilla;
				GetWindowRect(hwndEditor, &rcScintilla);
				RECT rcOutput;
				GetWindowRect(hwndOutput, &rcOutput);
				if (!PtInRect(&rcScintilla, ptCursor) && !PtInRect(&rcOutput, ptCursor)) {
					if (splitVertical)
						SetCursor(LoadCursor(NULL,IDC_SIZEWE));
					else
						SetCursor(LoadCursor(NULL,IDC_SIZENS));
					return TRUE;
				}
			}
			return DefWindowProc(hwndTide,iMessage,wParam,lParam);

		case WM_INITMENU:
			CheckMenus();
			break;

		case WM_DESTROY:
			if (hwndEditor)
				DestroyWindow(hwndEditor);
			hwndEditor = 0;
			if (hwndOutput)
				DestroyWindow(hwndOutput);
			hwndOutput = 0;
			PostQuitMessage(0);
			break;

		case WM_SETTINGCHANGE:
			dprintf("** Setting Changed\n");
			SendEditor(WM_SETTINGCHANGE, wParam, lParam);
			SendOutput(WM_SETTINGCHANGE, wParam, lParam);
			break;

		case WM_SYSCOLORCHANGE:
			dprintf("** Color Changed\n");
			SendEditor(WM_SYSCOLORCHANGE, wParam, lParam);
			SendOutput(WM_SYSCOLORCHANGE, wParam, lParam);
			break;

		case WM_ACTIVATEAPP:
			SendEditor(EM_HIDESELECTION, !wParam);
			break;

		case WM_ACTIVATE:
			SetFocus(hwndEditor);
			break;

		default:
			if (iMessage == RegisterWindowMessage(FINDMSGSTRING)) {
				HandleFindReplace();
				//dprintf("Find/replacemessage %x %x\n", wParam, lParam);
			} else {
				//dprintf("default wnd proc %x %d %d\n",iMessage, wParam, lParam);
				return DefWindowProc(hwndTide,iMessage,wParam,lParam);
			}
		}
		//dprintf("end wnd proc\n");
		return 0l;
	}

	LRESULT PASCAL TideWindow::TWndProc(
    	HWND hWnd,UINT iMessage,WPARAM wParam, LPARAM lParam) {
		dprintf("W:%x M:%d WP:%x L:%x\n", hWnd, iMessage, wParam, lParam);

		// Find C++ object associated with window.
		TideWindow *tide = (TideWindow *) GetWindowLong(hWnd,0);
		// tide will be zero if WM_CREATE not seen yet
		if (tide == 0) {
			if (iMessage == WM_CREATE) {
				LPCREATESTRUCT cs = reinterpret_cast<LPCREATESTRUCT>(lParam);
				tide = (TideWindow *) cs->lpCreateParams;
				tide->hwndTide = hWnd;
				SetWindowLong(hWnd, 0, reinterpret_cast<LONG>(tide));
				return tide->WndProc(iMessage, wParam, lParam);
			} else
				return DefWindowProc(hWnd, iMessage, wParam, lParam);
		} else
			return tide->WndProc(iMessage, wParam, lParam);
	}

	int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine,
                   	int nCmdShow) {
		OutputDebugString("TTT\r\n");
		dprintf("HW\n");

		dprintf("Command line is \n%s\n<<", lpszCmdLine);

		HACCEL hAccTable = LoadAccelerators(hInstance,"ACCELS");

		dprintf("HW\n");
		TideWindow::Register(hInstance);
#ifdef STATIC_BUILD
		dprintf("st\n");
		Scintilla_RegisterClasses(hInstance);
#else
		dprintf("sc\n");
		LoadLibrary("Scintilla.DLL");
#endif

		dprintf("ld\n");
		TideWindow MainWind(hInstance, lpszCmdLine);

		dprintf("ma\n");
		bool going = true;
		MSG msg;
		msg.wParam = 0;
		while (going) {
			going = GetMessage(&msg,NULL,0,0);
			if (currentDlg && going) {
				dprintf("DLG %x %x\n", currentDlg, msg.message);
				if (!IsDialogMessage(currentDlg, &msg)) {
					if (TranslateAccelerator(msg.hwnd,hAccTable,&msg) == 0) {
						TranslateMessage(&msg);
						DispatchMessage(&msg);
					}
				}
			} else if (going) {
				//if (TranslateAccelerator(msg.hwnd,hAccTable,&msg) == 0) {
				if (TranslateAccelerator(MainWind.GetWnd(),hAccTable,&msg) == 0) {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		}

		return msg.wParam;
	}
#endif
